"""Argument class and related functionality."""

import inspect
import json
import operator
import re
import sys
from collections.abc import Callable, Sequence
from contextlib import suppress
from functools import partial, reduce
from typing import TYPE_CHECKING, Any, get_args, get_origin

from attrs import define, field

from cyclopts._convert import (
    _validate_json_extra_keys,
    convert,
    instantiate_from_dict,
    token_count,
)
from cyclopts.annotations import (
    ITERABLE_TYPES,
    contains_hint,
    get_annotated_discriminator,
    is_attrs,
    is_dataclass,
    is_enum_flag,
    is_namedtuple,
    is_nonetype,
    is_pydantic,
    is_typeddict,
    is_union,
    resolve,
    resolve_annotated,
    resolve_optional,
)
from cyclopts.exceptions import (
    CoercionError,
    CycloptsError,
    MissingArgumentError,
    MixedArgumentError,
    RepeatArgumentError,
    ValidationError,
)
from cyclopts.field_info import (
    FieldInfo,
    _attrs_field_infos,
    _generic_class_field_infos,
    _pydantic_field_infos,
    _typed_dict_field_infos,
    get_field_infos,
    signature_parameters,
)
from cyclopts.parameter import ITERATIVE_BOOL_IMPLICIT_VALUE, Parameter
from cyclopts.token import Token
from cyclopts.utils import UNSET, grouper, is_builtin, parse_version

from .utils import (
    enum_flag_from_dict,
    get_choices_from_hint,
    missing_keys_factory,
    startswith,
)

if TYPE_CHECKING:
    from cyclopts.argument._collection import ArgumentCollection


@define(kw_only=True)
class Argument:
    """Encapsulates functionality and additional contextual information for parsing a parameter.

    An argument is defined as anything that would have its own entry in the help page.
    """

    tokens: list[Token] = field(factory=list)
    """
    List of :class:`.Token` parsed from various sources.
    Do not directly mutate; see :meth:`append`.
    """

    field_info: FieldInfo = field(factory=FieldInfo)
    """
    Additional information about the parameter from surrounding python syntax.
    """

    parameter: Parameter = field(factory=Parameter)
    """
    Fully resolved user-provided :class:`.Parameter`.
    """

    hint: Any = field(default=str, converter=resolve)
    """
    The type hint for this argument; may be different from :attr:`.FieldInfo.annotation`.
    """

    index: int | None = field(default=None)
    """
    Associated python positional index for argument.
    If ``None``, then cannot be assigned positionally.
    """

    keys: tuple[str, ...] = field(default=())
    """
    **Python** keys that lead to this leaf.

    ``self.parameter.name`` and ``self.keys`` can naively disagree!
    For example, a ``self.parameter.name="--foo.bar.baz"`` could be aliased to "--fizz".
    The resulting ``self.keys`` would be ``("bar", "baz")``.

    This is populated based on type-hints and class-structure, not ``Parameter.name``.

    .. code-block:: python

        from cyclopts import App, Parameter
        from dataclasses import dataclass
        from typing import Annotated

        app = App()


        @dataclass
        class User:
            id: int
            name: Annotated[str, Parameter(name="--fullname")]


        @app.default
        def main(user: User):
            pass


        for argument in app.assemble_argument_collection():
            print(f"name: {argument.name:16} hint: {str(argument.hint):16} keys: {str(argument.keys)}")

    .. code-block:: bash

        $ my-script
        name: --user.id        hint: <class 'int'>    keys: ('id',)
        name: --fullname       hint: <class 'str'>    keys: ('name',)
    """

    _value: Any = field(alias="value", default=UNSET)
    """
    Converted value from last :meth:`convert` call.
    This value may be stale if fields have changed since last :meth:`convert` call.
    :class:`.UNSET` if :meth:`convert` has not yet been called with tokens.
    """

    _accepts_keywords: bool = field(default=False, init=False, repr=False)

    _default: Any = field(default=None, init=False, repr=False)
    _lookup: dict[str, FieldInfo] = field(factory=dict, init=False, repr=False)

    children: "ArgumentCollection" = field(init=False, repr=False)
    """
    Collection of other :class:`Argument` that eventually culminate into the python variable represented by :attr:`field_info`.
    """

    _marked_converted: bool = field(default=False, init=False, repr=False)
    _mark_converted_override: bool = field(default=False, init=False, repr=False)

    _missing_keys_checker: Callable | None = field(default=None, init=False, repr=False)

    _internal_converter: Callable | None = field(default=None, init=False, repr=False)

    _enum_flag_type: Any | None = field(default=None, init=False, repr=False)

    def __attrs_post_init__(self):
        from cyclopts.argument._collection import ArgumentCollection

        self.children = ArgumentCollection()

        hint = resolve(self.hint)
        hints = get_args(hint) if is_union(hint) else (hint,)

        if self.parameter.count:
            # Perform type-annotation validation.
            resolved_hint = resolve_optional(hint)
            # Technically, bool is a subclass of int, so we need to explicitly check.
            if resolved_hint is bool or not (
                resolved_hint is int or (isinstance(resolved_hint, type) and issubclass(resolved_hint, int))
            ):
                raise ValueError(
                    f"Parameter(count=True) requires an int type hint, got {self.hint}. "
                    f"Use 'Annotated[int, Parameter(count=True)]' for counting flags."
                )

        if self.parameter.requires_equals and self.parameter.consume_multiple:
            raise ValueError(
                "Parameter(requires_equals=True) and Parameter(consume_multiple=...) cannot be used together. "
                "requires_equals enforces '--option=value' syntax, which is incompatible with "
                "consume_multiple's space-separated value consumption."
            )

        if not self.parse:
            # Validate that non-parsed parameters are keyword-only or have defaults
            is_keyword_only = self.field_info.kind is self.field_info.KEYWORD_ONLY
            has_default = self.field_info.default is not self.field_info.empty
            if not (is_keyword_only or has_default):
                raise ValueError(
                    f"Non-parsed parameter '{self.field_info.name}' must be a KEYWORD_ONLY function parameter "
                    "or have a default value."
                )
            return

        if self.parameter.accepts_keys is False:
            return

        for hint in hints:
            origin = get_origin(hint)
            hint_origin = {hint, origin}

            field_infos = get_field_infos(hint)
            if dict in hint_origin:
                self._accepts_keywords = True
                key_type, val_type = str, str
                args = get_args(hint)
                with suppress(IndexError):
                    key_type = args[0]
                    val_type = args[1]
                if key_type is not str:
                    raise TypeError('Dictionary type annotations must have "str" keys.')
                self._default = val_type
            elif is_typeddict(hint):
                self._missing_keys_checker = missing_keys_factory(_typed_dict_field_infos)
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif is_dataclass(hint):
                self._missing_keys_checker = missing_keys_factory(_generic_class_field_infos)
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif is_namedtuple(hint):
                self._missing_keys_checker = missing_keys_factory(_generic_class_field_infos)
                self._accepts_keywords = True
                if not hasattr(hint, "__annotations__"):
                    raise ValueError("Cyclopts cannot handle collections.namedtuple without type annotations.")
                self._update_lookup(field_infos)
            elif is_attrs(hint):
                self._missing_keys_checker = missing_keys_factory(_attrs_field_infos)
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif is_pydantic(hint):
                self._missing_keys_checker = missing_keys_factory(_pydantic_field_infos)
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif is_enum_flag(hint):
                self._enum_flag_type = hint
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif not is_builtin(hint) and field_infos:
                self._missing_keys_checker = missing_keys_factory(_generic_class_field_infos)
                self._accepts_keywords = True
                self._update_lookup(field_infos)
            elif self.parameter.accepts_keys is None:
                continue

            if self.parameter.accepts_keys is None:
                continue

            self._accepts_keywords = True
            self._missing_keys_checker = missing_keys_factory(_generic_class_field_infos)
            for i, field_info in enumerate(signature_parameters(hint.__init__).values()):
                if i == 0 and field_info.name == "self":
                    continue
                if field_info.kind is field_info.VAR_KEYWORD:
                    self._default = field_info.annotation
                else:
                    self._update_lookup({field_info.name: field_info})

    def _update_lookup(self, field_infos: dict[str, FieldInfo]):
        from typing import Literal

        discriminator = get_annotated_discriminator(self.field_info.annotation)

        for key, field_info in field_infos.items():
            if existing_field_info := self._lookup.get(key):
                if existing_field_info == field_info:
                    pass
                elif discriminator and discriminator in field_info.names and discriminator in existing_field_info.names:
                    existing_field_info.annotation = Literal[existing_field_info.annotation, field_info.annotation]
                    existing_field_info.default = FieldInfo.empty
                else:
                    raise NotImplementedError
            else:
                self._lookup[key] = field_info

    @property
    def value(self):
        """Converted value from last :meth:`convert` call.

        This value may be stale if fields have changed since last :meth:`convert` call.
        :class:`.UNSET` if :meth:`convert` has not yet been called with tokens.
        """
        return self._value

    @value.setter
    def value(self, val):
        if self._marked:
            self._mark_converted_override = True
        self._marked = True
        self._value = val

    @property
    def _marked(self):
        """If ``True``, then this node in the tree has already been converted and ``value`` has been populated."""
        return self._marked_converted | self._mark_converted_override

    @_marked.setter
    def _marked(self, value: bool):
        self._marked_converted = value

    @property
    def _accepts_arbitrary_keywords(self) -> bool:
        args = get_args(self.hint) if is_union(self.hint) else (self.hint,)
        return any(dict in (arg, get_origin(arg)) for arg in args)

    @property
    def show_default(self) -> bool | Callable[[Any], str]:
        """Show the default value on the help page."""
        if self.required:
            return False
        elif self.parameter.show_default is None:
            return self.field_info.default not in (None, self.field_info.empty)
        elif (self.field_info.default is self.field_info.empty) or not self.parameter.show_default:
            return False
        else:
            return self.parameter.show_default

    @property
    def _use_pydantic_type_adapter(self) -> bool:
        return bool(
            is_pydantic(self.hint)
            or (
                is_union(self.hint)
                and (
                    any(is_pydantic(x) for x in get_args(self.hint))
                    or get_annotated_discriminator(self.field_info.annotation)
                )
            )
        )

    def _type_hint_for_key(self, key: str):
        try:
            return self._lookup[key].annotation
        except KeyError:
            if self._default is None:
                raise
            return self._default

    def _should_attempt_json_dict(self, tokens: Sequence[Token | str] | None = None) -> bool:
        """When parsing, should attempt to parse the token(s) as json dict data."""
        if tokens is None:
            tokens = self.tokens
        if not tokens:
            return False
        value = tokens[0].value if isinstance(tokens[0], Token) else tokens[0]
        if not value.strip().startswith("{"):
            return False

        if self._accepts_keywords:
            if self.parameter.json_dict is not None:
                return self.parameter.json_dict
            if contains_hint(self.field_info.annotation, str):
                return False
            return True

        hint = resolve(self.hint)
        origin = get_origin(hint)
        if origin in ITERABLE_TYPES:
            args = get_args(hint)
            if args and args[0] is not str:
                return True

        return False

    def _should_attempt_json_list(
        self, tokens: Sequence[Token | str] | Token | str | None = None, keys: tuple[str, ...] = ()
    ) -> bool:
        """When parsing, should attempt to parse the token(s) as json list data."""
        if tokens is None:
            tokens = self.tokens
        if not tokens:
            return False
        _, consume_all = self.token_count(keys)
        if not consume_all:
            return False
        if isinstance(tokens, Token):
            value = tokens.value
        elif isinstance(tokens, str):
            value = tokens
        else:
            value = tokens[0].value if isinstance(tokens[0], Token) else tokens[0]
        if not value.strip().startswith("["):
            return False
        if self.parameter.json_list is not None:
            return self.parameter.json_list
        for arg in get_args(self.hint) or (str,):
            if contains_hint(arg, str):
                return False
        return True

    def match(
        self,
        term: str | int,
        *,
        transform: Callable[[str], str] | None = None,
        delimiter: str = ".",
    ) -> tuple[tuple[str, ...], Any]:
        """Match a name search-term, or a positional integer index.

        Raises
        ------
        ValueError
            If no match is found.

        Returns
        -------
        tuple[str, ...]
            Leftover keys after matching to this argument.
            Used if this argument accepts_arbitrary_keywords.
        Any
            Implicit value.
            :obj:`~.UNSET` if no implicit value is applicable.
        """
        if not self.parse:
            raise ValueError
        return (
            self._match_index(term)
            if isinstance(term, int)
            else self._match_name(term, transform=transform, delimiter=delimiter)
        )

    def _normalize_trailing_keys(self, trailing: tuple[str, ...]) -> tuple[str, ...]:
        """Map kebab-case segments back to their canonical Python field names.

        Walks the type hint segment-by-segment:

        * Dynamic ``dict`` keys pass through unchanged (advance to the value type).
        * Segments addressing a structured type (pydantic / dataclass / attrs /
          TypedDict / NamedTuple) are looked up in a ``{name_transform(name): name}``
          map built from the type's field_infos; on hit, the segment is replaced
          with the canonical name and the walk advances to that field's annotation.
        * On miss or when the hint is unwalkable (plain scalar, unresolved forward
          ref, etc.) remaining segments pass through unchanged — this preserves the
          existing raw-snake_case behavior as a backward-compat fallback.
        """
        name_transform = self.parameter.name_transform
        if name_transform is None or not trailing:
            return trailing

        # Seed from ``self.hint`` (not ``field_info.annotation``): for
        # ``**kwargs: SubConfig``, the annotation is ``SubConfig`` but the hint
        # is ``dict[str, SubConfig]`` — we need the wrapped form so the first
        # trailing segment is routed as a dict key rather than a field name.
        hint = resolve(self.hint)
        out: list[str] = []
        i = 0
        while i < len(trailing):
            segment = trailing[i]
            hint = resolve_optional(hint)

            if get_origin(hint) is dict:
                out.append(segment)
                args = get_args(hint)
                hint = args[1] if len(args) > 1 else str
                i += 1
                continue

            field_infos = {}
            try:
                field_infos = get_field_infos(hint)
            except Exception:
                pass
            if not field_infos:
                out.extend(trailing[i:])
                break

            # Build a kebab→canonical map from ``fi.names`` only.  Cyclopts's
            # field_info extractors populate ``names`` with exactly the names
            # the underlying library accepts (e.g. pydantic omits the python
            # name when ``populate_by_name=False``); trust that.
            name_map: dict[str, tuple[str, Any]] = {}
            for canonical_name, fi in field_infos.items():
                for alias in fi.names:
                    name_map.setdefault(name_transform(alias), (canonical_name, fi.annotation))

            match = name_map.get(segment)
            if match is None:
                out.extend(trailing[i:])
                break
            canonical_name, next_hint = match
            out.append(canonical_name)
            hint = resolve(next_hint)
            i += 1

        return tuple(out)

    def _match_name(
        self,
        term: str,
        *,
        transform: Callable[[str], str] | None = None,
        delimiter: str = ".",
    ) -> tuple[tuple[str, ...], Any]:
        """Check how well this argument matches a token keyword identifier.

        Parameters
        ----------
        term: str
            Something like "--foo"
        transform: Callable
            Function that converts the cyclopts Parameter name(s) into
            something that should be compared against ``term``.

        Raises
        ------
        ValueError
            If no match found.

        Returns
        -------
        tuple[str, ...]
            Leftover keys after matching to this argument.
            Used if this argument accepts_arbitrary_keywords.
        Any
            Implicit value.
        """
        if self.field_info.kind is self.field_info.VAR_KEYWORD:
            return self._normalize_trailing_keys(tuple(term.lstrip("-").split(delimiter))), UNSET

        trailing = term
        implicit_value = UNSET

        assert self.parameter.name
        for name in self.parameter.name:
            if transform:
                name = transform(name)
            if startswith(term, name):
                trailing = term[len(name) :]
                implicit_value = True if self.hint is bool or self.hint in ITERATIVE_BOOL_IMPLICIT_VALUE else UNSET
                if trailing:
                    if trailing[0] == delimiter:
                        trailing = trailing[1:]
                        break
                else:
                    return (), implicit_value
        else:
            hint = self._negatives_hint
            if is_union(hint):
                hints = get_args(hint)
            else:
                hints = (hint,)
            for hint in hints:
                hint = resolve_annotated(hint)
                double_break = False
                for name in self.parameter.get_negatives(hint):
                    if transform:
                        name = transform(name)
                    if startswith(term, name):
                        trailing = term[len(name) :]
                        if hint in ITERATIVE_BOOL_IMPLICIT_VALUE:
                            implicit_value = False
                        elif is_nonetype(hint) or hint is None:
                            implicit_value = None
                        else:
                            hint = resolve_optional(hint)
                            implicit_value = (get_origin(hint) or hint)()
                        if trailing:
                            if trailing[0] == delimiter:
                                trailing = trailing[1:]
                                double_break = True
                                break
                        else:
                            return (), implicit_value
                if double_break:
                    break
            else:
                raise ValueError

        if not self._accepts_arbitrary_keywords:
            raise ValueError

        return self._normalize_trailing_keys(tuple(trailing.split(delimiter))), implicit_value

    def _match_index(self, index: int) -> tuple[tuple[str, ...], Any]:
        if self.index is None:
            raise ValueError
        elif self.field_info.kind is self.field_info.VAR_POSITIONAL:
            if index < self.index:
                raise ValueError
        elif index != self.index:
            raise ValueError
        return (), UNSET

    def append(self, token: Token):
        """Safely add a :class:`Token`."""
        if not self.parse:
            raise ValueError

        if any(x.address == token.address for x in self.tokens):
            if self.parameter.allow_repeating is False:
                raise RepeatArgumentError(token=token)
            _, consume_all = self.token_count(token.keys)
            if self.parameter.allow_repeating is True:
                if not consume_all:
                    # "last wins" for scalar types — remove old tokens with same address
                    self.tokens = [x for x in self.tokens if x.address != token.address]
            elif not consume_all and not self.parameter.count:
                raise RepeatArgumentError(token=token)

        if self.tokens:
            if bool(token.keys) ^ any(x.keys for x in self.tokens):
                raise MixedArgumentError(argument=self)
        self.tokens.append(token)

    @property
    def has_tokens(self) -> bool:
        """This argument, or a child argument, has at least 1 parsed token."""  # noqa: D404
        return bool(self.tokens) or any(x.has_tokens for x in self.children)

    @property
    def children_recursive(self) -> "ArgumentCollection":
        from cyclopts.argument._collection import ArgumentCollection

        out = ArgumentCollection()
        for child in self.children:
            out.append(child)
            out.extend(child.children_recursive)
        return out

    def _convert_pydantic(self):
        if self.has_tokens:
            import pydantic

            unstructured_data = self._json()
            try:
                return pydantic.TypeAdapter(self.field_info.annotation).validate_python(unstructured_data)
            except pydantic.ValidationError as e:
                self._handle_pydantic_validation_error(e)
        else:
            return UNSET

    def _convert(self, converter: Callable | None = None):
        from cyclopts.argument._collection import update_argument_collection

        if self.parameter.converter:
            # Resolve string converters to methods on the type
            if isinstance(self.parameter.converter, str):
                converter = getattr(self.hint, self.parameter.converter)
            else:
                converter = self.parameter.converter
        elif converter is None:
            converter = partial(convert, name_transform=self.parameter.name_transform)

        assert converter is not None  # Ensure converter is set at this point

        def safe_converter(hint, tokens):
            if isinstance(tokens, dict):
                try:
                    return converter(hint, tokens)  # pyright: ignore
                except (AssertionError, ValueError, TypeError) as e:
                    raise CoercionError(msg=e.args[0] if e.args else None, argument=self, target_type=hint) from e
            else:
                try:
                    # Detect bound methods (classmethods/instance methods)
                    if inspect.ismethod(converter):
                        # Call with just tokens - cls/self already bound
                        return converter(tokens)  # pyright: ignore[reportCallIssue]
                    else:
                        # Regular function - pass type and tokens
                        return converter(hint, tokens)  # pyright: ignore[reportCallIssue]
                except (AssertionError, ValueError, TypeError) as e:
                    token = tokens[0] if len(tokens) == 1 else None
                    raise CoercionError(
                        msg=e.args[0] if e.args else None, argument=self, target_type=hint, token=token
                    ) from e

        if not self.parse:
            out = UNSET
        elif self.parameter.count:
            out = sum(token.implicit_value for token in self.tokens if token.implicit_value is not UNSET)
        elif not self.children:
            positional: list[Token] = []
            keyword = {}

            def expand_tokens(tokens):
                for token in tokens:
                    if self._should_attempt_json_list(token):
                        try:
                            parsed_json = json.loads(token.value)
                        except json.JSONDecodeError as e:
                            raise CoercionError(token=token, target_type=self.hint) from e

                        if not isinstance(parsed_json, list):
                            raise CoercionError(token=token, target_type=self.hint)

                        if not parsed_json:
                            yield token.evolve(value="", implicit_value=[])
                        else:
                            for element in parsed_json:
                                if element is None:
                                    yield token.evolve(value="", implicit_value=element)
                                elif isinstance(element, dict):
                                    yield token.evolve(value=json.dumps(element))
                                else:
                                    yield token.evolve(value=str(element))
                    else:
                        yield token

            expanded_tokens = list(expand_tokens(self.tokens))
            for token in expanded_tokens:
                resolved_hint = resolve_optional(self.hint)
                if token.implicit_value is not UNSET and isinstance(
                    token.implicit_value, get_origin(resolved_hint) or resolved_hint
                ):
                    assert len(expanded_tokens) == 1
                    return token.implicit_value

                if token.keys:
                    lookup = keyword
                    for key in token.keys[:-1]:
                        lookup = lookup.setdefault(key, {})
                    lookup.setdefault(token.keys[-1], []).append(token)
                else:
                    positional.append(token)

                if positional and keyword:  # pragma: no cover
                    raise MixedArgumentError(argument=self)

            if positional:
                if self.field_info and self.field_info.kind is self.field_info.VAR_POSITIONAL:
                    hint = get_args(self.hint)[0]
                    tokens_per_element, _ = self.token_count()
                    out = tuple(safe_converter(hint, values) for values in grouper(positional, tokens_per_element))
                else:
                    out = safe_converter(self.hint, tuple(positional))
            elif keyword:
                if self.field_info and self.field_info.kind is self.field_info.VAR_KEYWORD and not self.keys:
                    out = {key: safe_converter(get_args(self.hint)[1], value) for key, value in keyword.items()}
                else:
                    out = safe_converter(self.hint, keyword)
            elif self.required:
                raise MissingArgumentError(argument=self)
            else:
                return UNSET
        else:
            data = {}
            out = UNSET

            if self._enum_flag_type:
                out = self._enum_flag_type(0)

            if self._enum_flag_type and self.tokens:
                converted_flags = safe_converter(self._enum_flag_type, self.tokens)
                out |= reduce(operator.or_, converted_flags) if isinstance(converted_flags, list) else converted_flags

            if self._should_attempt_json_dict():
                while self.tokens:
                    token = self.tokens.pop(0)
                    try:
                        parsed_json = json.loads(token.value)
                    except json.JSONDecodeError as e:
                        raise CoercionError(token=token, target_type=self.hint) from e
                    _validate_json_extra_keys(parsed_json, self.hint, token)
                    update_argument_collection(
                        {self.name.lstrip("-"): parsed_json},
                        token.source,
                        self.children_recursive,
                        root_keys=(),
                        allow_unknown=False,
                    )

            if self._use_pydantic_type_adapter:
                return self._convert_pydantic()

            if self.tokens and not self._enum_flag_type:
                positional_tokens = [token for token in self.tokens if not token.keys]
                if positional_tokens:
                    return safe_converter(self.hint, tuple(positional_tokens))

            for child in self.children:
                assert len(child.keys) == (len(self.keys) + 1)
                if child.has_tokens:
                    data[child.keys[-1]] = child.convert_and_validate(converter=converter)
                elif child.required:
                    obj = data
                    for k in child.keys:
                        try:
                            obj = obj[k]
                        except Exception:
                            raise MissingArgumentError(argument=child) from None
                    child._marked = True

            self._run_missing_keys_checker(data)

            if self._enum_flag_type:
                out |= enum_flag_from_dict(self._enum_flag_type, data, self.parameter.name_transform)
                if not out:
                    out = UNSET
            elif data:
                out = instantiate_from_dict(self.hint, data)
            elif self.required:
                raise MissingArgumentError(argument=self)  # pragma: no cover
            else:
                out = UNSET

        return out

    def convert(self, converter: Callable | None = None):
        """Converts :attr:`tokens` into :attr:`value`.

        Parameters
        ----------
        converter: Callable | None
            Converter function to use. Overrides ``self.parameter.converter``

        Returns
        -------
        Any
            The converted data. Same as :attr:`value`.
        """
        if not self._marked:
            try:
                self.value = self._convert(converter=converter)
            except CoercionError as e:
                if e.argument is None:
                    e.argument = self
                if e.target_type is None:
                    e.target_type = self.hint
                raise
            except CycloptsError as e:
                if e.argument is None:
                    e.argument = self
                raise

        return self.value

    def validate(self, value):
        """Validates provided value.

        Parameters
        ----------
        value:
            Value to validate.

        Returns
        -------
        Any
            The converted data. Same as :attr:`value`.
        """
        assert isinstance(self.parameter.validator, tuple)

        # Only use pydantic validation if pydantic v2+ is available.
        # Pydantic v1 has an incompatible API (e.g. no TypeAdapter).
        if "pydantic" in sys.modules:
            import pydantic

            pydantic_version = parse_version(pydantic.__version__)
            if pydantic_version < (2,):
                pydantic = None
        else:
            pydantic = None

        def validate_pydantic(hint, val):
            if not pydantic:
                return
            if self._use_pydantic_type_adapter:
                return

            try:
                pydantic.TypeAdapter(hint).validate_python(val)
            except pydantic.ValidationError as e:
                self._handle_pydantic_validation_error(e)
            except pydantic.PydanticUserError:
                pass

        try:
            if not self.keys and self.field_info and self.field_info.kind is self.field_info.VAR_KEYWORD:
                hint = get_args(self.hint)[1]
                for validator in self.parameter.validator:
                    for val in value.values():
                        validator(hint, val)
                validate_pydantic(dict[str, self.field_info.annotation], value)
            elif self.field_info and self.field_info.kind is self.field_info.VAR_POSITIONAL:
                hint = get_args(self.hint)[0]
                for validator in self.parameter.validator:
                    for val in value:
                        validator(hint, val)
                validate_pydantic(tuple[self.field_info.annotation, ...], value)
            else:
                for validator in self.parameter.validator:
                    validator(self.hint, value)
                validate_pydantic(self.field_info.annotation, value)
        except (AssertionError, ValueError, TypeError) as e:
            raise ValidationError(exception_message=e.args[0] if e.args else "", argument=self) from e

    def convert_and_validate(self, converter: Callable | None = None):
        """Converts and validates :attr:`tokens` into :attr:`value`.

        Parameters
        ----------
        converter: Callable | None
            Converter function to use. Overrides ``self.parameter.converter``

        Returns
        -------
        Any
            The converted data. Same as :attr:`value`.
        """
        val = self.convert(converter=converter)
        if val is not UNSET:
            self.validate(val)
        elif self.field_info.default is not FieldInfo.empty:
            self.validate(self.field_info.default)
        return val

    def token_count(self, keys: tuple[str, ...] = ()):
        """The number of string tokens this argument consumes.

        Parameters
        ----------
        keys: tuple[str, ...]
            The **python** keys into this argument.
            If provided, returns the number of string tokens that specific
            data type within the argument consumes.

        Returns
        -------
        int
            Number of string tokens to create 1 element.
        consume_all: bool
            :obj:`True` if this data type is iterable.
        """
        if self.parameter.count:
            return 0, False

        # Check for explicit n_tokens override
        # This applies to values at any level: root values (keys=()) or nested values (keys=(...))
        # For example, **kwargs: Annotated[str, Parameter(n_tokens=2)] means each kwarg value needs 2 tokens
        if self.parameter.n_tokens is not None:
            if self.parameter.n_tokens == -1:
                return 1, True
            else:
                # Determine consume_all based on the hint at the requested level
                # by recursively calling token_count on the hint
                if len(keys) > 1:
                    hint = self._default
                elif len(keys) == 1:
                    hint = self._type_hint_for_key(keys[0])
                else:
                    hint = self.hint

                # Recursively call token_count to get the consume_all behavior
                # We ignore the token count from the recursive call and use our explicit n_tokens
                _, consume_all_from_type = token_count(hint)
                return self.parameter.n_tokens, consume_all_from_type

        if len(keys) > 1:
            hint = self._default
        elif len(keys) == 1:
            hint = self._type_hint_for_key(keys[0])
        else:
            hint = self.hint
            if self._enum_flag_type and not keys:
                return 1, True
        tokens_per_element, consume_all = token_count(hint)
        return tokens_per_element, consume_all

    @property
    def _negatives_hint(self):
        # Mirrors ``field_info.annotation`` but substitutes ``type(default)`` when
        # there is no annotation, so an unannotated ``foo=False`` is treated as
        # ``bool`` for negative-flag purposes. Unlike ``self.hint``, this preserves
        # ``Optional`` / unions, which ``negative_none`` depends on.
        hint = self.field_info.annotation
        if hint is inspect.Parameter.empty or resolve(hint) is Any:
            default = self.field_info.default
            if default is not inspect.Parameter.empty and default is not None:
                hint = type(default)
        return resolve_annotated(hint)

    @property
    def negatives(self):
        """Negative flags from :meth:`.Parameter.get_negatives`."""
        return self.parameter.get_negatives(self._negatives_hint)

    @property
    def name(self) -> str:
        """The **first** provided name this argument goes by."""
        return self.names[0]

    @property
    def names(self) -> tuple[str, ...]:
        """Names the argument goes by (both positive and negative)."""
        import itertools

        assert isinstance(self.parameter.name, tuple)
        return tuple(itertools.chain(self.parameter.name, self.negatives))

    def env_var_split(self, value: str, delimiter: str | None = None) -> list[str]:
        """Split a given value with :meth:`.Parameter.env_var_split`."""
        return self.parameter.env_var_split(self.hint, value, delimiter=delimiter)

    @property
    def show(self) -> bool:
        """Show this argument on the help page.

        If an argument has child arguments, don't show it on the help-page.
        Returns False for arguments that won't be parsed (including underscore-prefixed params).
        """
        if self.children:
            return False
        if self.parameter.show is not None:
            # User explicitly set show
            return self.parameter.show
        # Default to whether this argument is parsed
        return self.parse

    @property
    def parse(self) -> bool:
        """Whether this argument should be parsed from CLI tokens.

        If ``Parameter.parse`` is a regex pattern, parse if the pattern matches
        the field name; otherwise don't parse.
        """
        if self.parameter.parse is None:
            return True
        if isinstance(self.parameter.parse, re.Pattern):
            return bool(self.parameter.parse.search(self.field_info.name))
        return bool(self.parameter.parse)

    @property
    def required(self) -> bool:
        """Whether or not this argument requires a user-provided value."""
        if self.parameter.required is None:
            return self.field_info.required
        else:
            return self.parameter.required

    def is_positional_only(self) -> bool:
        return self.field_info.is_positional_only

    def is_var_positional(self) -> bool:
        return self.field_info.kind == self.field_info.VAR_POSITIONAL

    def is_flag(self) -> bool:
        """Check if this argument is a flag (consumes no CLI tokens).

        Flags are arguments that don't consume command-line tokens after the option name.
        They typically have implicit values (e.g., `--verbose` for bool, `--no-items` for list).

        Returns
        -------
        bool
            True if the argument consumes zero tokens from the command line.

        Examples
        --------
        >>> from cyclopts import Parameter
        >>> bool_arg = Argument(hint=bool, parameter=Parameter(name="--verbose"))
        >>> bool_arg.is_flag()
        True
        >>> str_arg = Argument(hint=str, parameter=Parameter(name="--name"))
        >>> str_arg.is_flag()
        False
        """
        return self.token_count() == (0, False)

    def get_choices(self, force: bool = False) -> tuple[str, ...] | None:
        """Extract completion choices from type hint.

        Extracts choices from Literal types, Enum types, and Union types containing them.
        Respects the Parameter.show_choices setting unless force=True.

        Parameters
        ----------
        force : bool
            If True, return choices even when show_choices=False.
            Used by shell completion to always provide choices.

        Returns
        -------
        tuple[str, ...] | None
            Tuple of choice strings if choices exist and should be shown, None otherwise.

        Examples
        --------
        >>> argument = Argument(hint=Literal["dev", "staging", "prod"], parameter=Parameter(show_choices=True))
        >>> argument.get_choices()
        ('dev', 'staging', 'prod')
        >>> argument = Argument(hint=Literal["dev", "staging", "prod"], parameter=Parameter(show_choices=False))
        >>> argument.get_choices()  # Returns None for help text
        >>> argument.get_choices(force=True)  # Returns choices for completion
        ('dev', 'staging', 'prod')
        """
        if not force and not self.parameter.show_choices:
            return None
        choices = get_choices_from_hint(self.hint, self.parameter.name_transform)
        return tuple(choices) if choices else None

    def _json(self) -> dict:
        """Convert argument to be json-like for pydantic.

        All values will be str/list/dict. JSON-serialized strings (from sources
        like config files or environment variables) are deserialized back to their
        original dict/list structure.
        """
        out = {}
        if self._accepts_keywords:
            for token in self.tokens:
                node = out
                for key in token.keys[:-1]:
                    node = node.setdefault(key, {})
                node[token.keys[-1]] = token.value if token.implicit_value is UNSET else token.implicit_value
        for child in self.children:
            child._marked = True
            if not child.has_tokens:
                continue
            keys = child.keys[len(self.keys) :]
            if child._accepts_keywords:
                result = child._json()
                if result:
                    out[keys[0]] = result
            elif (get_origin(child.hint) or child.hint) in ITERABLE_TYPES:
                for token in child.tokens:
                    if token.implicit_value is not UNSET:
                        out.setdefault(keys[-1], []).extend(token.implicit_value)
                    else:
                        value = token.value
                        # Deserialize JSON strings (from update_argument_collection) back to dict/list
                        if isinstance(value, str) and value.strip() and value.strip()[0] in ("{", "["):
                            try:
                                value = json.loads(value)
                            except json.JSONDecodeError:
                                pass
                        out.setdefault(keys[-1], []).append(value)
            else:
                token = child.tokens[0]
                out[keys[0]] = token.value if token.implicit_value is UNSET else token.implicit_value
        return out

    def _run_missing_keys_checker(self, data):
        if not self._missing_keys_checker or (not self.required and not data):
            return
        if not (missing_keys := self._missing_keys_checker(self, data)):
            return
        missing_key = missing_keys[0]
        keys = self.keys + (missing_key,)
        missing_arguments = self.children.filter_by(keys_prefix=keys)
        if missing_arguments:
            raise MissingArgumentError(argument=missing_arguments[0])
        else:
            missing_description = self.field_info.names[0] + "->" + "->".join(keys)
            raise ValueError(
                f'Required field "{missing_description}" is not accessible by Cyclopts; possibly due to conflicting POSITIONAL/KEYWORD requirements.'
            )

    def _handle_pydantic_validation_error(self, exc):
        import pydantic

        error = exc.errors()[0]
        if error["type"] == "missing":
            loc = error["loc"]

            # Pydantic includes list indices in loc for list-element errors
            # (e.g. ("animals", 0, "dog", "name")). Cyclopts doesn't model list
            # items as individual Arguments, so no prefix-match can correctly
            # map — fall through to the native pydantic error, which shows the
            # full nested path.
            if not any(isinstance(part, int) for part in loc):
                candidate = tuple(loc)
                while candidate:
                    missing_arguments = self.children_recursive.filter_by(keys_prefix=self.keys + candidate)
                    # An Argument that already has tokens cannot be "missing";
                    # the stripping heuristic has wandered into a populated sibling.
                    missing_arguments = [a for a in missing_arguments if not a.tokens]
                    if missing_arguments:
                        raise MissingArgumentError(argument=missing_arguments[0]) from exc
                    if len(candidate) == 1:
                        break
                    # For discriminated unions pydantic prepends the discriminator
                    # value (e.g. loc=("cat", "rainbow")). Strip leading elements
                    # until we find a real child Argument.
                    candidate = candidate[1:]
            # Fall through to ValidationError.

        if isinstance(exc, pydantic.ValidationError):
            raise ValidationError(exception_message=str(exc), argument=self) from exc
        else:
            raise exc



"""Argument and ArgumentCollection classes for CLI parsing."""

from cyclopts.token import Token

from ._argument import Argument
from ._collection import (
    ArgumentCollection,
    _resolve_groups_from_callable,
    update_argument_collection,
)
from .utils import get_choices_from_hint, resolve_parameter_name

__all__ = [
    "Argument",
    "ArgumentCollection",
    "Token",
    "_resolve_groups_from_callable",
    "get_choices_from_hint",
    "resolve_parameter_name",
    "update_argument_collection",
]



from textwrap import dedent

from cyclopts.config._yaml import Yaml


def test_config_yaml(tmp_path):
    fn = tmp_path / "test.yaml"
    fn.write_text(
        dedent(
            """\
            foo:
                key1: foo1
                key2: foo2

                function1:
                    key1: bar1
                    key2: bar2
            """
        )
    )
    config = Yaml(fn)
    assert config.config == {
        "foo": {
            "key1": "foo1",
            "key2": "foo2",
            "function1": {
                "key1": "bar1",
                "key2": "bar2",
            },
        }
    }



"""Tests for cyclopts.docs.base helpers."""

from cyclopts.docs.base import apply_usage_name


def test_apply_usage_name_none_returns_chain_unchanged():
    chain = ["cli", "files", "cp"]
    assert apply_usage_name(chain, None) is chain


def test_apply_usage_name_empty_chain_returns_single_element_list():
    assert apply_usage_name([], "uv run cli") == ["uv run cli"]


def test_apply_usage_name_replaces_root_only():
    chain = ["cli", "files", "cp"]
    assert apply_usage_name(chain, "uv run cli") == ["uv run cli", "files", "cp"]


def test_apply_usage_name_does_not_mutate_input():
    chain = ["cli", "files"]
    apply_usage_name(chain, "uv run cli")
    assert chain == ["cli", "files"]


def test_apply_usage_name_empty_string_drops_root():
    assert apply_usage_name(["cli", "files"], "") == ["files"]
    assert apply_usage_name(["cli"], "") == []
    assert apply_usage_name([], "") == []



from dataclasses import dataclass
from typing import Annotated
from unittest.mock import Mock

import pytest

from cyclopts import Parameter, ValidationError
from cyclopts.exceptions import CoercionError
from cyclopts.token import Token


@pytest.fixture
def validator():
    return Mock()


def test_custom_converter(app, assert_parse_args):
    def custom_converter(type_, tokens):
        return 2 * int(tokens[0].value)

    @app.default
    def foo(age: Annotated[int, Parameter(converter=custom_converter)]):
        pass

    assert_parse_args(foo, "5", age=10)


def test_custom_converter_dict(app, assert_parse_args):
    def custom_converter(type_, tokens):
        return {k: 2 * int(v[0].value) for k, v in tokens.items()}

    @app.default
    def foo(*, color: Annotated[dict[str, int], Parameter(converter=custom_converter)]):
        pass

    assert_parse_args(foo, "--color.red 5 --color.green 10", color={"red": 10, "green": 20})


def test_custom_converter_user_value_error_single_token(app):
    def custom_converter(type_, tokens):
        raise ValueError

    @app.default
    def foo(age: Annotated[int, Parameter(converter=custom_converter)]):
        pass

    with pytest.raises(CoercionError) as e:
        app("5", exit_on_error=False)
    assert str(e.value) == 'Invalid value for "AGE": unable to convert "5" into int.'


def test_custom_converter_user_value_error_multi_token(app):
    def custom_converter(type_, tokens):
        raise ValueError

    @app.default
    def foo(age: Annotated[tuple[int, int], Parameter(converter=custom_converter)]):
        pass

    with pytest.raises(CoercionError) as e:
        app("5 6", exit_on_error=False)
    assert str(e.value) == 'Invalid value for "--age": unable to convert value to tuple[int, int].'


def test_custom_converter_user_value_error_with_message(app):
    def custom_converter(type_, tokens):
        raise ValueError("Some user-provided message.")

    @app.default
    def foo(age: Annotated[int, Parameter(converter=custom_converter)]):
        pass

    with pytest.raises(CoercionError) as e:
        app("5", exit_on_error=False)
    assert str(e.value) == "Some user-provided message."


def test_custom_converter_user_kwargs_error(app):
    def custom_converter(type_, tokens):
        raise ValueError

    @app.default
    def foo(**kwargs: Annotated[int, Parameter(converter=custom_converter)]):
        pass

    with pytest.raises(CoercionError) as e:
        app("--foo 5", exit_on_error=False)
    assert str(e.value) == 'Invalid value for "--foo": unable to convert "5" into int.'


def test_custom_converter_user_kwargs_error_with_message(app):
    def custom_converter(type_, tokens):
        raise ValueError("Some user-provided message.")

    @app.default
    def foo(**kwargs: Annotated[int, Parameter(converter=custom_converter)]):
        pass

    with pytest.raises(CoercionError) as e:
        app("--foo 5", exit_on_error=False)
    assert str(e.value) == "Invalid value for --foo: Some user-provided message."


def test_custom_validator_positional_or_keyword(app, assert_parse_args, validator):
    @app.default
    def foo(age: Annotated[int, Parameter(validator=validator)]):
        pass

    assert_parse_args(foo, "10", age=10)
    validator.assert_called_once_with(int, 10)


def test_custom_validator_var_keyword(app, assert_parse_args, validator):
    @app.default
    def foo(**age: Annotated[int, Parameter(validator=validator)]):
        pass

    assert_parse_args(foo, "--age=10", age=10)
    validator.assert_called_once_with(int, 10)


def test_custom_validator_var_positional(app, assert_parse_args, validator):
    @app.default
    def foo(*age: Annotated[int, Parameter(validator=validator)]):
        pass

    assert_parse_args(foo, "10", 10)
    validator.assert_called_once_with(int, 10)


def test_custom_validators(app, assert_parse_args):
    def lower_bound(type_, value):
        if value <= 0:
            raise ValueError("An unreasonable age was entered.")

    def upper_bound(type_, value):
        if value > 150:
            raise ValueError("An unreasonable age was entered.")

    @app.default
    def foo(age: Annotated[int, Parameter(validator=[lower_bound, upper_bound])]):
        pass

    assert_parse_args(foo, "10", 10)

    with pytest.raises(ValidationError):
        app.parse_args("0", print_error=False, exit_on_error=False)

    with pytest.raises(ValidationError):
        app.parse_args("200", print_error=False, exit_on_error=False)


def test_custom_converter_and_validator(app, assert_parse_args, validator):
    def custom_validator(type_, value):
        if not (0 < value < 150):
            raise ValueError("An unreasonable age was entered.")

    def custom_converter(type_, tokens):
        return 2 * int(tokens[0].value)

    @app.default
    def foo(age: Annotated[int, Parameter(converter=custom_converter, validator=validator)]):
        pass

    assert_parse_args(foo, "5", 10)

    validator.assert_called_once_with(int, 10)


def test_custom_validator_on_default_signature_value(app, validator):
    @app.default
    def foo(age: Annotated[int, Parameter(validator=validator)] = -1):
        pass

    app.parse_args("", print_error=False, exit_on_error=False)
    validator.assert_called_once_with(int, -1)


def test_custom_command_validator(app, assert_parse_args):
    validator = Mock()

    @app.default(validator=validator)
    def foo(a: int, b: int, c: int):
        pass

    assert_parse_args(foo, "1 2 3", 1, 2, 3)
    validator.assert_called_once_with(a=1, b=2, c=3)


def test_custom_converter_inside_class(app, mocker):
    converter = mocker.Mock(return_value=5)

    @Parameter(name="*")
    @dataclass
    class Config:
        foo: Annotated[int, Parameter(converter=converter)]

    @app.default
    def default(config: Config):
        pass

    app("bar")

    converter.assert_called_once_with(int, (Token(value="bar", source="cli"),))



from functools import partial
from typing import Annotated, Any

import pytest

from cyclopts import Parameter
from cyclopts.exceptions import (
    ArgumentOrderError,
    CoercionError,
    MissingArgumentError,
    RepeatArgumentError,
    UnknownCommandError,
    UnknownOptionError,
    UnusedCliTokensError,
)
from cyclopts.group import Group


def test_parse_known_args(app):
    @app.command
    def foo(a: int, b: int):
        pass

    command, _, unused_tokens, ignored = app.parse_known_args("foo 1 2 --bar 100")
    assert ignored == {}
    assert command == foo
    assert unused_tokens == ["--bar", "100"]


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1 2 3",
        "foo 1 2 --c=3",
        "foo --a 1 --b 2 --c 3",
        "foo --c 3 --b=2 --a 1",
    ],
)
def test_basic_1(app, cmd_str, assert_parse_args):
    @app.command
    def foo(a: int, b: int, c: int):
        pass

    assert_parse_args(foo, cmd_str, 1, 2, 3)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1 2 3 --d 10 --some-flag",
        "foo --some-flag 1 --b=2 --c 3 --d 10",
        "foo 1 2 --some-flag 3 --d 10",
    ],
)
def test_basic_2(app, cmd_str, assert_parse_args):
    @app.command
    def foo(a: int, b: int, c: int, d: int = 5, *, some_flag: bool = False):
        pass

    assert_parse_args(foo, cmd_str, 1, 2, 3, d=10, some_flag=True)


def test_functools_partial_default(app, assert_parse_args):
    def foo(a: int, b: int, c: int):
        pass

    foo_partial = partial(foo, c=3)
    app.default(foo_partial)

    assert_parse_args(foo_partial, "1 2", a=1, b=2)


def test_functools_partial_command(app, assert_parse_args):
    def foo(a: int, b: int, c: int):
        pass

    foo_partial = partial(foo, c=3)
    app.command(foo_partial)

    assert_parse_args(foo_partial, "foo 1 2", a=1, b=2)


def test_basic_allow_hyphen_or_underscore(app, assert_parse_args):
    @app.default
    def default(foo_bar):
        pass

    assert_parse_args(default, "--foo-bar=bazz", "bazz")
    assert_parse_args(default, "--foo_bar=bazz", "bazz")


def test_out_of_order_mixed_positional_or_keyword(app, assert_parse_args):
    @app.command
    def foo(a, b, c):
        pass

    with pytest.raises(ArgumentOrderError):
        app.parse_args("foo --b=5 1 2", print_error=False, exit_on_error=False)


def test_command_rename(app, assert_parse_args):
    @app.command(name="bar")
    def foo():
        pass

    assert_parse_args(foo, "bar")


def test_command_delete(app, assert_parse_args):
    @app.command
    def foo():
        pass

    del app["foo"]

    with pytest.raises(UnknownCommandError):
        assert_parse_args(foo, "foo")


def test_command_multiple_alias(app, assert_parse_args):
    @app.command(name=["bar", "baz"])
    def foo():
        pass

    assert_parse_args(foo, "bar")
    assert_parse_args(foo, "baz")


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo --age 10",
        "foo --duration 10",
        "foo -a 10",
    ],
)
def test_multiple_names(app, cmd_str, assert_parse_args):
    @app.command
    def foo(age: Annotated[int, Parameter(name=["--age", "--duration", "-a"])]):
        pass

    assert_parse_args(foo, cmd_str, age=10)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo --age 10",
        "foo --duration 10",
        "foo -a 10",
    ],
)
def test_alias(app, cmd_str, assert_parse_args):
    @app.command
    def foo(age: Annotated[int, Parameter(alias=["--duration", "-a"])]):
        pass

    assert_parse_args(foo, cmd_str, age=10)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo --age 10",
        "foo --duration 10",
        "foo -a 10",
    ],
)
def test_name_and_alias(app, cmd_str, assert_parse_args):
    """Weird use-case, but should be handled."""

    @app.command
    def foo(age: Annotated[int, Parameter(name="--age", alias=["--duration", "-a"])]):
        pass

    assert_parse_args(foo, cmd_str, age=10)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "--job-name foo",
        "-j foo",
    ],
)
def test_short_name_j(app, cmd_str, assert_parse_args):
    """
    "-j" previously didn't work as a short-name because it's a valid complex value.

    https://github.com/BrianPugh/cyclopts/issues/328
    """

    @app.default
    def main(
        *,
        job_name: Annotated[str, Parameter(name=["--job-name", "-j"], negative=False)],
    ):
        pass

    assert_parse_args(main, cmd_str, job_name="foo")


def test_short_flag_combining(app, assert_parse_args):
    @app.default
    def main(
        foo: Annotated[bool, Parameter(name=("--foo", "-f"))] = False,
        bar: Annotated[bool, Parameter(name=("--bar", "-b"))] = False,
        my_list: Annotated[list | None, Parameter(negative=("--empty-my-list", "-e"))] = None,
    ):
        pass

    # Note: ``my_list`` is explicitly getting an empty list.
    assert_parse_args(main, "-bfe", foo=True, bar=True, my_list=[])


def test_short_flag_combining_unknown_flag(app, assert_parse_args):
    @app.default
    def main(
        foo: Annotated[bool, Parameter(name=("--foo", "-f"))] = False,
        bar: Annotated[bool, Parameter(name=("--bar", "-b"))] = False,
    ):
        pass

    with pytest.raises(UnknownOptionError):
        # The flag "-e" is unknown
        app("-be", exit_on_error=False)


def test_short_flag_combining_with_short_option(app, assert_parse_args):
    @app.default
    def main(
        *,
        foo: Annotated[bool, Parameter(name=("--foo", "-f"))] = False,
        bar: Annotated[str, Parameter(name=("--bar", "-b"))],
    ):
        pass

    # With GNU-style support, -fb is now valid: -f (flag) + -b (needs value from next token)
    # Since no next token exists, should raise MissingArgumentError
    with pytest.raises(MissingArgumentError):
        app("-fb", exit_on_error=False)

    # But -fb with a value should work
    assert_parse_args(main, "-fb value", foo=True, bar="value")


def test_short_integer_flag(app, assert_parse_args):
    @app.default
    def main(
        foo: Annotated[int, Parameter(name=("--foo", "-f"))],
        *,
        bar: Annotated[bool, Parameter(name=("--bar", "-1"))],
    ):
        pass

    assert_parse_args(main, "-1 -2", -2, bar=True)


@pytest.mark.parametrize(
    "cmd_str,expected",
    [
        ("-9", -9),
        ("-10", -10),  # Multi-digit negatives should NOT be treated as combined short flags
        ("-100", -100),
        ("-3.14", -3.14),
        ("-1e5", -1e5),
        ("-1.5e-3", -1.5e-3),
    ],
)
def test_negative_number_not_combined_short_flags(app, cmd_str, expected, assert_parse_args):
    """Negative numbers should be parsed as values, not as combined short flags.

    Regression test for issue where -10 was incorrectly interpreted as combined
    short flags -1 and -0, while -9 worked because len("-9") == 2 doesn't trigger
    combined flag parsing (which requires len > 2).
    """

    @app.default
    def main(value: float):
        pass

    assert_parse_args(main, cmd_str, expected)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo --age 10",
        "foo --duration 10",
        "foo -a 10",
    ],
)
def test_multiple_names_no_hyphen(app, cmd_str, assert_parse_args):
    @app.command
    def foo(age: Annotated[int, Parameter(name=["age", "duration", "-a"])]):
        pass

    assert_parse_args(foo, cmd_str, age=10)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_optional_nonrequired_implicit_coercion(app, cmd_str, annotated, assert_parse_args):
    """
    For a union without an explicit coercion, the first non-None type annotation
    should be used. In this case, it's ``int``.
    """
    if annotated:

        @app.command
        def foo(a: Annotated[int | None, Parameter(help="help for a")] = None):
            pass

    else:

        @app.command
        def foo(a: int | None = None):
            pass

    assert_parse_args(foo, cmd_str, 1)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_optional_nonrequired_implicit_coercion_python310_syntax(app, cmd_str, annotated, assert_parse_args):
    """
    For a union without an explicit coercion, the first non-None type annotation
    should be used. In this case, it's ``int``.
    """
    if annotated:

        @app.command
        def foo(a: Annotated[int | None, Parameter(help="help for a")] = None):  # pyright: ignore
            pass

    else:

        @app.command
        def foo(a: int | None = None):  # pyright: ignore
            pass

    assert_parse_args(foo, cmd_str, 1)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "--foo val1 --foo val2",
    ],
)
def test_exception_repeat_argument(app, cmd_str):
    @app.default
    def default(foo: str):
        pass

    with pytest.raises(RepeatArgumentError):
        app.parse_args(cmd_str, print_error=False, exit_on_error=False)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "--foo val1 --foo val2",
    ],
)
def test_exception_repeat_argument_kwargs(app, cmd_str):
    @app.default
    def default(**kwargs: str):
        pass

    with pytest.raises(RepeatArgumentError):
        app.parse_args(cmd_str, print_error=False, exit_on_error=False)


@pytest.mark.parametrize("type_hint", [list[str], str])
def test_allow_repeating_false(app, type_hint):
    @app.default
    def default(foo: Annotated[type_hint, Parameter(allow_repeating=False)]):  # pyright: ignore[reportInvalidTypeForm]
        pass

    with pytest.raises(RepeatArgumentError):
        app.parse_args("--foo a --foo b", print_error=False, exit_on_error=False)


def test_allow_repeating_false_consume_multiple(app, assert_parse_args):
    @app.default
    def default(foo: Annotated[list[str], Parameter(allow_repeating=False, consume_multiple=True)]):
        pass

    assert_parse_args(default, "--foo a b c", foo=["a", "b", "c"])


def test_allow_repeating_false_consume_multiple_repeated(app):
    @app.default
    def default(foo: Annotated[list[str], Parameter(allow_repeating=False, consume_multiple=True)]):
        pass

    with pytest.raises(RepeatArgumentError):
        app.parse_args("--foo a --foo b", print_error=False, exit_on_error=False)


@pytest.mark.parametrize(
    "consume_multiple,cmd_str,expected",
    [
        (False, "--foo a --foo b", ["a", "b"]),
        (True, "--foo a --foo b", ["a", "b"]),
        (True, "--foo a b --foo c d", ["a", "b", "c", "d"]),
    ],
)
def test_allow_repeating_true_list(app, assert_parse_args, consume_multiple, cmd_str, expected):
    @app.default
    def default(foo: Annotated[list[str], Parameter(allow_repeating=True, consume_multiple=consume_multiple)]):
        pass

    assert_parse_args(default, cmd_str, foo=expected)


@pytest.mark.parametrize(
    "cmd_str,expected",
    [
        ("--foo a --foo b", "b"),
        ("--foo a --foo b --foo c", "c"),
    ],
)
def test_allow_repeating_true_scalar(app, assert_parse_args, cmd_str, expected):
    @app.default
    def default(foo: Annotated[str, Parameter(allow_repeating=True)]):
        pass

    assert_parse_args(default, cmd_str, foo=expected)


def test_exception_unused_token(app):
    @app.default
    def default(foo: str):
        pass

    with pytest.raises(UnusedCliTokensError):
        app.parse_args("foo bar", print_error=False, exit_on_error=False)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_bind_no_hint_no_default(app, cmd_str, annotated, assert_parse_args):
    """Parameter with no type hint & no default should be treated as a ``str``."""
    if annotated:

        @app.command
        def foo(a: Annotated[Any, Parameter(help="help for a")]):  # pyright: ignore[reportRedeclaration]
            pass

    else:

        @app.command
        def foo(a):  # pyright: ignore[reportRedeclaration]
            pass

    assert_parse_args(foo, cmd_str, "1")


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_bind_no_hint_none_default(app, cmd_str, annotated, assert_parse_args):
    """Parameter with no type hint & ``None`` default should be treated as a ``str``."""
    if annotated:

        @app.command
        def foo(a: Annotated[Any, Parameter(help="help for a")] = None):  # pyright: ignore[reportRedeclaration]
            pass

    else:

        @app.command
        def foo(a=None):  # pyright: ignore[reportRedeclaration]
            pass

    assert_parse_args(foo, cmd_str, "1")


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_bind_no_hint_typed_default(app, cmd_str, annotated, assert_parse_args):
    """Parameter with no type hint & typed default should be treated as a ``type(default)``."""
    if annotated:

        @app.command
        def foo(a: Annotated[Any, Parameter(help="help for a")] = 5):  # pyright: ignore[reportRedeclaration]
            pass

    else:

        @app.command
        def foo(a=5):  # pyright: ignore[reportRedeclaration]
            pass

    assert_parse_args(foo, cmd_str, 1)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "foo 1",
        "foo --a=1",
        "foo --a 1",
    ],
)
@pytest.mark.parametrize("annotated", [False, True])
def test_bind_any_hint(app, cmd_str, annotated, assert_parse_args):
    """The ``Any`` type hint should be treated as a ``str``."""
    if annotated:

        @app.command
        def foo(a: Annotated[Any, Parameter(help="help for a")] = None):
            pass

    else:

        @app.command
        def foo(a: Any = None):
            pass

    assert_parse_args(foo, cmd_str, "1")


@pytest.mark.parametrize(
    "cmd_str",
    [
        "1",
        "0b1",
        "0x01",
        "1.0",
        "0.9",
    ],
)
def test_bind_int_advanced(app, cmd_str, assert_parse_args):
    @app.default
    def foo(a: int):
        pass

    assert_parse_args(foo, cmd_str, 1)


def test_bind_int_advanced_coercion_error(app):
    @app.default
    def foo(a: int):
        pass

    with pytest.raises(CoercionError):
        app.parse_args("foo", exit_on_error=False)


@pytest.mark.parametrize(
    "value",
    [
        "10000000000000001",
        "10000000000000011",
        "100000000000000111234",
    ],
)
def test_bind_large_int(app, assert_parse_args, value):
    """Test that large integers are parsed correctly without precision loss.

    Reproduces bug: https://github.com/BrianPugh/cyclopts/issues/581
    """

    @app.default
    def foo(a: int):
        pass

    assert_parse_args(foo, value, int(value))


def test_bind_override_app_groups(app):
    g_commands = Group("Custom Commands")
    g_arguments = Group("Custom Arguments")
    g_parameters = Group("Custom Parameters")

    @app.command(group_commands=g_commands, group_arguments=g_arguments, group_parameters=g_parameters)
    def foo():
        pass

    assert app["foo"].group_commands == g_commands
    assert app["foo"].group_arguments == g_arguments
    assert app["foo"].group_parameters == g_parameters


def test_bind_version(app, capsys):
    app.version = "1.2.3"
    actual_command, actual_bind, ignored = app.parse_args("--version")
    assert ignored == {}
    assert actual_command == app.version_print

    actual_command(*actual_bind.args, **actual_bind.kwargs)
    captured = capsys.readouterr()
    assert captured.out == "1.2.3\n"


def test_bind_version_factory(app, capsys):
    app.version = lambda: "1.2.3"
    actual_command, actual_bind, ignored = app.parse_args("--version")
    assert ignored == {}
    assert actual_command == app.version_print

    actual_command(*actual_bind.args, **actual_bind.kwargs)
    captured = capsys.readouterr()
    assert captured.out == "1.2.3\n"


@pytest.mark.parametrize(
    "cmd_str_e",
    [
        ("foo 1 2 3", MissingArgumentError),
        ("foo 1 2", MissingArgumentError),
    ],
)
def test_missing_keyword_argument(app, cmd_str_e):
    cmd_str, e = cmd_str_e

    @app.command
    def foo(a: int, b: int, c: int, *, d: int):
        pass

    with pytest.raises(e):
        app.parse_args(cmd_str, print_error=False, exit_on_error=False)


@pytest.mark.parametrize(
    "cmd_str",
    [
        "1 -- --2 3 4",
        "-- 1 --2 3 4",
        "--c=3 4 -- 1 --2",
        "--c 3 4 -- 1 --2",
    ],
)
def test_default_double_hyphen_end_of_options_delimiter(app, cmd_str, assert_parse_args):
    @app.default
    def foo(a: int, b: str, c: tuple[int, int]):
        pass

    assert_parse_args(foo, cmd_str, 1, "--2", (3, 4))


def test_disabled_double_hyphen_end_of_options_delimiter_from_app(app, assert_parse_args):
    app.end_of_options_delimiter = ""

    @app.default
    def foo(a: int, b: Annotated[str, Parameter(allow_leading_hyphen=True)], c: tuple[int, int]):
        pass

    assert_parse_args(foo, "1 -- 3 4", 1, "--", (3, 4))


def test_disabled_double_hyphen_end_of_options_delimiter_from_parse_args(app, assert_parse_args_config):
    @app.default
    def foo(a: int, b: Annotated[str, Parameter(allow_leading_hyphen=True)], c: tuple[int, int]):
        pass

    assert_parse_args_config({"end_of_options_delimiter": ""}, foo, "1 -- 3 4", 1, "--", (3, 4))


def test_end_of_options_delimiter_from_parse_args(app, assert_parse_args):
    app.end_of_options_delimiter = "AND"

    @app.default
    def foo(a: int, b: str, c: tuple[int, int]):
        pass

    assert_parse_args(foo, "1 AND --2 3 4", 1, "--2", (3, 4))


def test_end_of_options_delimiter_override(app, assert_parse_args_config):
    app.end_of_options_delimiter = "AND"  # This gets overridden

    @app.default
    def foo(a: int, b: str, c: tuple[int, int]):
        pass

    assert_parse_args_config({"end_of_options_delimiter": "DELIMIT"}, foo, "1 DELIMIT --2 3 4", 1, "--2", (3, 4))


# ============================================================================
# GNU-style combined short options tests
# Tests for the feature where short options can be combined with their values
# in a single token, similar to GNU getopt behavior.
#
# Examples:
#     -o9         →  -o with value "9"
#     -iuroot     →  -i (flag), -u with value "root"
#     -fvbfile    →  -f (flag), -v (flag), -b with value "file"
# ============================================================================


def test_single_short_option_with_attached_value(app, assert_parse_args):
    """Test that -o9 is equivalent to -o 9."""

    @app.default
    def main(output: Annotated[str, Parameter(name=("-o", "--output"))]):
        pass

    # Traditional syntax should still work
    assert_parse_args(main, "-o 9", output="9")
    assert_parse_args(main, "-o=9", output="9")

    # NEW: Attached value should work
    assert_parse_args(main, "-o9", output="9")


def test_single_short_option_with_attached_string_value(app, assert_parse_args):
    """Test that -ofile.txt works for string values."""

    @app.default
    def main(output: Annotated[str, Parameter(name=("-o", "--output"))]):
        pass

    # Traditional syntax
    assert_parse_args(main, "-o file.txt", output="file.txt")

    # NEW: Attached string value
    assert_parse_args(main, "-ofile.txt", output="file.txt")
    assert_parse_args(main, "-oroot", output="root")


def test_combined_flags_before_option_with_value(app, assert_parse_args):
    """Test sudo -iuroot style: flags followed by option with attached value."""

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
        interactive: Annotated[bool, Parameter(name=("-i", "--interactive"))] = False,
    ):
        pass

    # Traditional syntax
    assert_parse_args(main, "-i -u root", interactive=True, user="root")

    # NEW: GNU-style combined
    assert_parse_args(main, "-iuroot", interactive=True, user="root")


def test_multiple_flags_before_option_with_value(app, assert_parse_args):
    """Test multiple flags combined with an option that takes a value."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("-o", "--output"))],
        force: Annotated[bool, Parameter(name=("-f", "--force"))] = False,
        verbose: Annotated[bool, Parameter(name=("-v", "--verbose"))] = False,
    ):
        pass

    # Traditional syntax
    assert_parse_args(main, "-f -v -o file.txt", force=True, verbose=True, output="file.txt")

    # NEW: GNU-style combined
    assert_parse_args(main, "-fvofile.txt", force=True, verbose=True, output="file.txt")
    assert_parse_args(main, "-vfofile.txt", verbose=True, force=True, output="file.txt")


def test_option_with_value_at_end_consumes_next_token(app, assert_parse_args):
    """Test that -fu (without attached value) consumes next token."""

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
        force: Annotated[bool, Parameter(name=("-f", "--force"))] = False,
    ):
        pass

    # When -u is at the end without attached value, next token is the value
    assert_parse_args(main, "-fu root", force=True, user="root")


def test_option_with_value_in_middle_stops_processing(app, assert_parse_args):
    """Test that once we hit an option with a value, rest is consumed as the value."""

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
        force: Annotated[bool, Parameter(name=("-f", "--force"))] = False,
        verbose: Annotated[bool, Parameter(name=("-v", "--verbose"))] = False,
    ):
        pass

    # -fuvroot should be: -f (flag), -u with value "vroot"
    # The 'v' is NOT treated as a flag, it's part of the value
    assert_parse_args(main, "-fuvroot", force=True, user="vroot")


def test_combined_short_options_numeric_value(app, assert_parse_args):
    """Test numeric values attached to short options."""

    @app.default
    def main(
        port: Annotated[int, Parameter(name=("-p", "--port"))],
        verbose: Annotated[bool, Parameter(name=("-v", "--verbose"))] = False,
    ):
        pass

    # Traditional syntax
    assert_parse_args(main, "-v -p 8080", verbose=True, port=8080)

    # NEW: Attached numeric value
    assert_parse_args(main, "-vp8080", verbose=True, port=8080)
    assert_parse_args(main, "-p8080", port=8080)


def test_gnu_backward_compatibility_all_flags(app, assert_parse_args):
    """Test that combining only flags still works (no regression)."""

    @app.default
    def main(
        force: Annotated[bool, Parameter(name=("-f", "--force"))] = False,
        verbose: Annotated[bool, Parameter(name=("-v", "--verbose"))] = False,
        quiet: Annotated[bool, Parameter(name=("-q", "--quiet"))] = False,
    ):
        pass

    # This should continue to work as before
    assert_parse_args(main, "-fvq", force=True, verbose=True, quiet=True)
    assert_parse_args(main, "-qvf", quiet=True, verbose=True, force=True)


def test_gnu_backward_compatibility_space_separated(app, assert_parse_args):
    """Ensure traditional space-separated syntax still works."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("-o", "--output"))],
        user: Annotated[str, Parameter(name=("-u", "--user"))],
    ):
        pass

    # Traditional syntax must still work
    assert_parse_args(main, "-o file.txt -u root", output="file.txt", user="root")


def test_gnu_backward_compatibility_equals_syntax(app, assert_parse_args):
    """Ensure equals syntax still works."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("-o", "--output"))],
    ):
        pass

    # Equals syntax must still work
    assert_parse_args(main, "-o=file.txt", output="file.txt")
    assert_parse_args(main, "--output=file.txt", output="file.txt")


def test_empty_string_value_attached(app, assert_parse_args):
    """Test edge case: can we have an empty attached value?"""

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
        force: Annotated[bool, Parameter(name=("-f", "--force"))] = False,
    ):
        pass

    # -fu with no characters after should consume next token
    # This is actually the same as test_option_with_value_at_end_consumes_next_token
    assert_parse_args(main, "-fu root", force=True, user="root")


def test_counting_parameter_combined(app, assert_parse_args):
    """Test that counting parameters work in combinations."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("-o", "--output"))],
        verbose: Annotated[int, Parameter(name=("-v", "--verbose"), count=True)] = 0,
    ):
        pass

    # Multiple v's for verbosity levels, then option with value
    assert_parse_args(main, "-vvvofile.txt", verbose=3, output="file.txt")
    assert_parse_args(main, "-vvofile.txt", verbose=2, output="file.txt")


def test_only_option_no_flags(app, assert_parse_args):
    """Test single option with attached value (simplest case)."""

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
    ):
        pass

    assert_parse_args(main, "-uroot", user="root")
    assert_parse_args(main, "-u root", user="root")


def test_long_option_should_not_be_split(app, assert_parse_args):
    """Test that long options (--) are NOT affected by this feature."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("--output",))],
    ):
        pass

    # Long options should still require space or equals
    assert_parse_args(main, "--output file.txt", output="file.txt")
    assert_parse_args(main, "--output=file.txt", output="file.txt")

    # --outputfile.txt should NOT be split into --output with value file.txt
    # It should be treated as an unknown option "--outputfile.txt"
    # (This maintains current behavior)


def test_hyphenated_value_attached(app, assert_parse_args):
    """Test that hyphenated values work when attached."""

    @app.default
    def main(
        output: Annotated[str, Parameter(name=("-o", "--output"))],
    ):
        pass

    # Value containing hyphens
    assert_parse_args(main, "-omy-file.txt", output="my-file.txt")
    assert_parse_args(main, "-o--weird-value", output="--weird-value")


def test_special_characters_in_attached_value(app, assert_parse_args):
    """Test special characters in attached values."""

    @app.default
    def main(
        pattern: Annotated[str, Parameter(name=("-p", "--pattern"))],
    ):
        pass

    # Special characters should be preserved
    assert_parse_args(main, "-p*.txt", pattern="*.txt")
    assert_parse_args(main, "-p[a-z]+", pattern="[a-z]+")
    # Note: For short options with GNU-style attachment, = is part of the value
    # Use -p "file=value" or --pattern=file for splitting on =
    assert_parse_args(main, "-pfile=value", pattern="file=value")


def test_multiple_options_only_first_gets_attached_value(app, assert_parse_args):
    """Test behavior when multiple value-taking options are combined.

    Only the first option can have an attached value. Once we hit a value-taking
    option, everything after is the value.
    """

    @app.default
    def main(
        user: Annotated[str, Parameter(name=("-u", "--user"))],
        password: Annotated[str | None, Parameter(name=("-p", "--password"))] = None,
    ):
        pass

    # -uprootpass should be: -u with value "prootpass"
    # The 'p' is NOT treated as another option, it's part of the value
    assert_parse_args(main, "-uprootpass", user="prootpass")

    # To set both, use separate options or space
    assert_parse_args(main, "-uroot -ppass", user="root", password="pass")


@pytest.mark.parametrize(
    "cmd,expected_output",
    [
        ("-o9", "9"),
        ("-o99", "99"),
        ("-o123abc", "123abc"),
        ("-oabc123", "abc123"),
    ],
)
def test_various_attached_values(app, assert_parse_args, cmd, expected_output):
    """Parametrized test for various value formats."""

    @app.default
    def main(output: Annotated[str, Parameter(name=("-o", "--output"))]):
        pass

    assert_parse_args(main, cmd, output=expected_output)


def test_single_char_option_with_negative_value(app, assert_parse_args):
    """Test that -o-5 treats -5 as the value for -o."""

    @app.default
    def main(
        offset: Annotated[int, Parameter(name=("-o", "--offset"))],
    ):
        pass

    # -o-5 should be: -o with value "-5"
    # This tests that the "-5" part is treated as a value, not as a separate option
    assert_parse_args(main, "-o-5", offset=-5)


def test_positional_only_list_interleaved_with_keywords_error(app):
    """Positional tokens after keyword args should not be consumed by a positional-only list.

    Regression test for issue #763.
    """

    @app.default
    def main(foo: list[str] | None = None, /, *, bar: int = 0, baz: int = 0):
        pass

    with pytest.raises(UnusedCliTokensError):
        app.parse_args("a b c --bar 8 --baz 10 d", print_error=False, exit_on_error=False)


def test_positional_only_list_then_keywords(app):
    """Positional tokens before keywords should work fine.

    Regression test for issue #763.
    """

    @app.default
    def main(foo: list[str] | None = None, /, *, bar: int = 0, baz: int = 0):
        pass

    _, actual_bind, _ = app.parse_args("a b c d --bar 8 --baz 10", print_error=False, exit_on_error=False)
    assert actual_bind.args == (["a", "b", "c", "d"],)
    assert actual_bind.kwargs == {"bar": 8, "baz": 10}


def test_positional_only_list_keywords_first(app):
    """Keywords before positional tokens should work fine.

    Regression test for issue #763.
    """

    @app.default
    def main(foo: list[str] | None = None, /, *, bar: int = 0, baz: int = 0):
        pass

    _, actual_bind, _ = app.parse_args("--bar 8 --baz 10 a b c d", print_error=False, exit_on_error=False)
    assert actual_bind.args == (["a", "b", "c", "d"],)
    assert actual_bind.kwargs == {"bar": 8, "baz": 10}


def test_positional_only_list_interleaved_with_delimiter(app):
    """Tokens after -- should still be consumed even with interleaving before the delimiter.

    Regression test for issue #763.
    """

    @app.default
    def main(foo: list[str] | None = None, /, *, bar: int = 0):
        pass

    _, actual_bind, _ = app.parse_args("a b --bar 8 -- d e", print_error=False, exit_on_error=False)
    assert actual_bind.args == (["a", "b", "d", "e"],)
    assert actual_bind.kwargs == {"bar": 8}


def test_positional_only_list_and_scalar_interleaved_error(app):
    """Multiple positional-only params (list + scalar) should also reject interleaving.

    Regression test for issue #763.
    """
    from pathlib import Path

    @app.default
    def main(inputs: list[str], output: Path, /, *, verbose: bool = False):
        pass

    # Non-interleaved: should work
    _, actual_bind, _ = app.parse_args("a b c out.csv --verbose", print_error=False, exit_on_error=False)
    assert actual_bind.args == (["a", "b", "c"], Path("out.csv"))
    assert actual_bind.kwargs == {"verbose": True}

    # Interleaved: should error
    with pytest.raises(UnusedCliTokensError):
        app.parse_args("a b --verbose c out.csv", print_error=False, exit_on_error=False)



from textwrap import dedent

import pytest

from cyclopts import CycloptsError
from cyclopts.exceptions import MissingArgumentError, UnknownCommandError


@pytest.fixture
def mock_get_function_info(mocker):
    mocker.patch("cyclopts.exceptions._get_function_info", return_value=("FILENAME", 100))


def test_runtime_exception_not_enough_tokens(app, console, mock_get_function_info):
    @app.default
    def foo(a: tuple[int, int, int]):
        pass

    with console.capture() as capture, pytest.raises(CycloptsError):
        app(["1", "2"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == (
        "╭─ Error ────────────────────────────────────────────────────────────╮\n"
        '│ Parameter "--a" requires 3 positional arguments. Only got 2.       │\n'
        "╰────────────────────────────────────────────────────────────────────╯\n"
    )

    with console.capture() as capture, pytest.raises(CycloptsError):
        app(["1", "2"], exit_on_error=False, error_console=console, verbose=True)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ MissingArgumentError                                               │
        │ Function defined in file "FILENAME", line 100:                     │
        │     foo(a: tuple[int, int, int])                                   │
        │ Root Input Tokens: ['1', '2']                                      │
        │ Parameter "--a" requires 3 positional arguments. Only got 2.       │
        │ Parsed: ['1', '2'].                                                │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )


def test_runtime_exception_missing_parameter(app, console):
    @app.default
    def foo(a, b, c):
        pass

    with console.capture() as capture, pytest.raises(CycloptsError):
        app(["1", "2"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == (
        "╭─ Error ────────────────────────────────────────────────────────────╮\n"
        '│ Parameter "--c" requires an argument.                              │\n'
        "╰────────────────────────────────────────────────────────────────────╯\n"
    )


def test_runtime_exception_bad_command(app, console):
    with console.capture() as capture, pytest.raises(UnknownCommandError):
        app(["bad-command", "123"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == (
        "╭─ Error ────────────────────────────────────────────────────────────╮\n"
        '│ Unknown command "bad-command".                                     │\n'
        "╰────────────────────────────────────────────────────────────────────╯\n"
    )


def test_runtime_exception_bad_command_recommend(app, console):
    @app.command
    def mad_command():
        pass

    with console.capture() as capture, pytest.raises(UnknownCommandError):
        app(["bad-command", "123"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ Unknown command "bad-command". Did you mean "mad-command"?         │
        │ Available commands: mad-command.                                   │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )


def test_runtime_exception_bad_command_recommend_no_show(app, console):
    """If a command is hidden, do not show recommendations for it."""

    @app.command(show=False)
    def mad_command():  # "mad-command" should not be recommended, and not show up as an available command.
        pass

    @app.command
    def other_command():
        pass

    with console.capture() as capture, pytest.raises(UnknownCommandError):
        app(["bad-command", "123"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ Unknown command "bad-command". Did you mean "other-command"?       │
        │ Available commands: other-command.                                 │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )


def test_runtime_exception_bad_command_list_ellipsis(app, console):
    def cmd():
        pass

    app.command(name="cmd1")(cmd)
    app.command(name="cmd2")(cmd)
    app.command(name="cmd3")(cmd)
    app.command(name="cmd4")(cmd)
    app.command(name="cmd5")(cmd)
    app.command(name="cmd6")(cmd)
    app.command(name="cmd7")(cmd)
    app.command(name="cmd8")(cmd)
    app.command(name="cmd9")(cmd)

    with console.capture() as capture, pytest.raises(UnknownCommandError):
        app(["cmd", "123"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ Unknown command "cmd". Did you mean "cmd9"? Available commands:    │
        │ cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, ...                │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )


def test_runtime_exception_bad_parameter_recommend(app, console):
    @app.command
    def some_command(*, foo: int):
        pass

    with console.capture() as capture, pytest.raises(MissingArgumentError):
        app(["some-command", "--boo", "123"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ Command "some-command" parameter "--foo" requires an argument. Did │
        │ you mean "--foo" instead of "--boo"?                               │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )


def test_runtime_exception_repeat_arguments(app, console):
    @app.default
    def foo(a):
        pass

    with console.capture() as capture, pytest.raises(CycloptsError):
        app(["--a=1", "--a=2"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == (
        "╭─ Error ────────────────────────────────────────────────────────────╮\n"
        "│ Parameter --a specified multiple times.                            │\n"
        "╰────────────────────────────────────────────────────────────────────╯\n"
    )


def test_runtime_exception_missing_tuple(app, console):
    """
    A bug was found where the "Did you mean" would be inappropriately displayed
    when insufficient tokens were supplied to a tuple type.

    https://github.com/BrianPugh/cyclopts/issues/443
    """

    @app.default
    def main(
        *,
        network_delay: tuple[int, int] | None = None,
    ):
        pass

    with console.capture() as capture, pytest.raises(MissingArgumentError):
        app(["--network-delay", "1"], exit_on_error=False, error_console=console)

    actual = capture.get()
    assert actual == dedent(
        """\
        ╭─ Error ────────────────────────────────────────────────────────────╮
        │ Parameter "--network-delay" requires 2 positional arguments. Only  │
        │ got 1.                                                             │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )



__all__ = [
    "all_or_none",
    "LimitedChoice",
    "MutuallyExclusive",
    "mutually_exclusive",
    "Number",
    "Path",
]

from cyclopts.validators._group import LimitedChoice, MutuallyExclusive, all_or_none, mutually_exclusive
from cyclopts.validators._number import Number
from cyclopts.validators._path import Path



"""Type definitions and utilities for documentation generation."""

from typing import Literal

# All accepted format values (canonical and aliases)
DocFormat = Literal["markdown", "md", "html", "htm", "rst", "rest", "restructuredtext"]

# Canonical format values only
CanonicalDocFormat = Literal["markdown", "html", "rst"]

# Map all aliases to their canonical format
# Also used for suffix lookups by stripping the leading period
FORMAT_ALIASES: dict[str, CanonicalDocFormat] = {
    "md": "markdown",
    "markdown": "markdown",
    "html": "html",
    "htm": "html",
    "rst": "rst",
    "rest": "rst",
    "restructuredtext": "rst",
}


def normalize_format(format_value: str) -> CanonicalDocFormat:
    """Normalize format aliases to standard format names.

    Parameters
    ----------
    format_value : str
        The format string to normalize (case-insensitive).

    Returns
    -------
    CanonicalDocFormat
        The canonical format name.

    Raises
    ------
    ValueError
        If the format is not recognized.
    """
    format_lower = format_value.lower()
    canonical_format = FORMAT_ALIASES.get(format_lower)

    if canonical_format is None:
        raise ValueError(
            f'Unsupported format "{format_value}". Supported formats: {", ".join(sorted(FORMAT_ALIASES.keys()))}'
        )

    return canonical_format



from collections import namedtuple
from typing import NamedTuple


def test_bind_typing_named_tuple(app, assert_parse_args):
    class Employee(NamedTuple):
        name: str
        id: int = 3

    @app.command
    def foo(user: Employee):
        pass

    assert_parse_args(
        foo,
        'foo --user.name="John Smith" --user.id=100',
        Employee(name="John Smith", id=100),
    )

    assert_parse_args(
        foo,
        'foo "John Smith" 100',
        Employee(name="John Smith", id=100),
    )


def test_bind_typing_named_tuple_var_positional(app, assert_parse_args):
    class Employee(NamedTuple):
        name: str
        id: int

    @app.command
    def foo(*users: Employee):
        pass

    assert_parse_args(
        foo,
        'foo "John Smith" 100 "Mary Jones" 200',
        Employee(name="John Smith", id=100),
        Employee(name="Mary Jones", id=200),
    )


def test_bind_collections_named_tuple(app, assert_parse_args):
    # All fields will be strings since cyclopts doesn't know the types.
    Employee = namedtuple("Employee", ["name", "id"])

    @app.command
    def foo(user: Employee):
        pass

    assert_parse_args(
        foo,
        'foo --user.name="John Smith" --user.id=100',
        Employee(name="John Smith", id="100"),
    )



import arguably


@arguably.command
def ec2__start_instances(*instances):
    """Start instances.

    Args:
        *instances: {instance}s to start
    """
    for inst in instances:
        print(f"Starting {inst}")


@arguably.command
def ec2__stop_instances(*instances):
    """Stop instances.

    Args:
        *instances: {instance}s to stop
    """
    for inst in instances:
        print(f"Stopping {inst}")


if __name__ == "__main__":
    arguably.run()

from cyclopts import App

app = App()
ec2 = app.command(App(name="ec2"))


@ec2.command
def start_instances(*instances):
    """Start instances.

    Args:
        *instances: {instance}s to start
    """
    for inst in instances:
        print(f"Starting {inst}")


@ec2.command
def stop_instances(*instances):
    """Stop instances.

    Args:
        *instances: {instance}s to stop
    """
    for inst in instances:
        print(f"Stopping {inst}")


if __name__ == "__main__":
    app()



from typing import NewType

import pytest

from cyclopts._convert import token_count


def test_new_type_str(app, assert_parse_args):
    CustomStr = NewType("CustomStr", str)

    @app.default
    def main(a: CustomStr):
        pass

    assert_parse_args(main, "foo", CustomStr("foo"))


def test_new_type_token_count_str(app, assert_parse_args):
    CustomStr = NewType("CustomStr", str)
    assert (1, False) == token_count(CustomStr)


@pytest.mark.parametrize(
    "cmd, expected",
    [
        ("foo", ["foo"]),
        ("--a foo", ["foo"]),
        ("foo bar", ["foo", "bar"]),
        ("--a foo --a bar", ["foo", "bar"]),
        ("foo bar baz", ["foo", "bar", "baz"]),
        ("--a foo --a bar --a baz", ["foo", "bar", "baz"]),
    ],
)
def test_new_type_token_count_list_str(app, assert_parse_args, cmd, expected):
    CustomStr = NewType("CustomStr", str)

    @app.default
    def main(a: list[CustomStr]):
        pass

    assert_parse_args(main, cmd, expected)



"""Cross-shell behavioral contract tests.

Most tests in this directory assert on the **text of the generated completion
script** (string containment). That is fast and catches regressions in the
generators, but it does not answer the real question: *when the user presses
TAB, does the shell actually offer the right suggestions?*

This file runs the same behavioral scenarios across bash, zsh, and fish via
the non-interactive drivers in ``conftest.py``. Any per-shell divergence
surfaces immediately as a single failing parametrization.

Assertions use subset / exclusion (not exact equality) because shells
legitimately differ in ordering and in whether helper entries like ``--help``
surface at every depth.

Known divergences (intentional, not bugs):

* **Naked-TAB at a no-arg subcommand**: bash returns nothing; zsh's
  ``_arguments`` always lists ``--help`` / ``--version``. Bash's
  ``cur == -*`` gate is the right convention — typing space-TAB usually
  means "show me commands or positional values," not "show me flags," and
  mixing flags into that suggestion list clutters the UI. The current
  per-shell behavior is locked in by ``test_naked_tab_*`` in
  ``test_bash.py`` / ``test_zsh.py``.
* **``--opt=value`` form** is now supported in both shells (Group C), but
  the cross-shell scenario list below excludes it because each shell
  reaches it through a different code path and the *space* form alone
  exercises the binding logic well enough.
"""

from __future__ import annotations

import pytest

from .apps import (
    app_basic,
    app_disabled_negative,
    app_enum,
    app_list_path,
    app_multiple_positionals,
    app_negative,
    app_nested,
    app_positional_literal,
    app_positional_path,
    app_three_positionals,
    app_two_iterables,
)

# Each scenario:
#   id            - human-readable name (used as pytest id)
#   app           - App object
#   prog_name     - program name (matches what the generator was invoked with)
#   partial       - the command-line fragment the user has typed
#   contains      - suggestions that MUST appear (subset check)
#   excludes      - suggestions that MUST NOT appear (exclusion check)
SCENARIOS = [
    # 1. Root-level subcommands listed.
    (
        "root-subcommands",
        app_basic,
        "basic",
        "basic ",
        {"deploy"},
        set(),
    ),
    # 2. Subcommand prefix filters correctly.
    (
        "subcommand-prefix",
        app_basic,
        "basic",
        "basic de",
        {"deploy"},
        set(),
    ),
    # 3. Long-flag prefix.
    (
        "flag-prefix",
        app_basic,
        "basic",
        "basic --v",
        {"--verbose"},
        set(),
    ),
    # 4. All top-level options when user has typed just "-".
    (
        "all-flags",
        app_basic,
        "basic",
        "basic -",
        {"--verbose", "--count"},
        set(),
    ),
    # 5. Negative form of a bool flag.
    (
        "negative-flag",
        app_negative,
        "negapp",
        "negapp --no-",
        {"--no-colors"},
        set(),
    ),
    # 6. Literal choices after an option that takes a value.
    (
        "literal-option-value",
        app_basic,
        "basic",
        "basic deploy --env ",
        {"dev", "staging", "prod"},
        set(),
    ),
    # 7. Literal positional (positional-only).
    (
        "literal-positional",
        app_positional_literal,
        "poslit",
        "poslit command ",
        {"foo", "bar", "baz"},
        set(),
    ),
    # 8. Position-aware multi-positional: second arg has different choices.
    (
        "multi-positional-second",
        app_multiple_positionals,
        "multipos",
        "multipos command-multi red ",
        {"cat", "dog"},
        {"red", "blue"},
    ),
    # 9. Nested subcommand discovery.
    (
        "nested-subcommand",
        app_nested,
        "nested",
        "nested config ",
        {"get", "set"},
        set(),
    ),
    # 10. Path-typed positional triggers file completion (shell-specific).
    #     We can't assert specific filenames, so we only assert the driver
    #     returns *some* non-empty result when a file exists in cwd.
    (
        "path-positional",
        app_positional_path,
        "pathpos",
        "pathpos process ",
        None,  # special-cased in the test body
        set(),
    ),
    # 11. Enum values.
    (
        "enum-value",
        app_enum,
        "enumapp",
        "enumapp --speed ",
        {"fast", "slow"},
        set(),
    ),
    # 12. Disabled-negatives + list-of-literal.
    (
        "list-of-literal",
        app_disabled_negative,
        "disabledneg",
        "disabledneg build --param ",
        {"apple", "banana", "cherry"},
        set(),
    ),
    # 13. --help is always available at the root.
    (
        "help-at-root",
        app_basic,
        "basic",
        "basic --h",
        {"--help"},
        set(),
    ),
    # 14. --opt=value form. Each shell reaches value-completion through a
    # different code path (bash: COMP_WORDBREAK splits on '=' so the script
    # walks back through ``$prev``; zsh: ``_arguments`` parses ``--env=`` as
    # a single token; fish: ``complete -C`` does the right thing natively).
    # Excluded historically out of caution — this is exactly where per-shell
    # divergence regressions hide.
    (
        "equals-form-option-value",
        app_basic,
        "basic",
        "basic deploy --env=p",
        {"prod"},
        set(),
    ),
    # 15. --help works at a nested command depth, not just the root.
    (
        "help-at-nested-depth",
        app_nested,
        "nested",
        "nested config --h",
        {"--help"},
        set(),
    ),
    # 16. Position-aware multi-positional: third slot has its own choices.
    # Scenario #8 only verifies the second slot — this catches off-by-one
    # regressions deeper in the cycle.
    (
        "multi-positional-third",
        app_three_positionals,
        "multipos3",
        "multipos3 command-multi3 red cat ",
        {"small", "large"},
        {"red", "blue", "cat", "dog"},
    ),
    # 17. Iterable positional (``list[Path]``) keeps offering completions
    # past the first slot — a regression net for the rest-owner logic.
    # Path completion is shell-specific, so we only assert non-empty (same
    # convention as scenario #10).
    (
        "iterable-positional-rest",
        app_list_path,
        "listpath",
        "listpath a.txt ",
        None,
        set(),
    ),
    # 18. Two iterable positionals back-to-back. The "first iterable wins
    # the rest spec" rule (commits 1bc65e8 / 3302dc7) prevents a doubled
    # rest spec from being emitted. With the bug present, scripts either
    # fail to parse or silently produce wrong completions at the rest slot.
    (
        "two-iterables-rest-owner",
        app_two_iterables,
        "twoiter",
        "twoiter collect a.txt b.txt ",
        None,
        set(),
    ),
]


@pytest.fixture(params=["bash", "zsh", "fish"])
def shell_tester_factory(request):
    """Pick the right tester factory for this shell parametrization.

    Resolves the target shell's fixture lazily via ``getfixturevalue`` so a
    missing shell (e.g. fish absent locally) skips only its own
    parametrizations, not every row in the matrix.
    """
    return request.getfixturevalue(f"{request.param}_tester")


@pytest.mark.parametrize(
    ("scenario_id", "app", "prog_name", "partial", "contains", "excludes"),
    SCENARIOS,
    ids=[s[0] for s in SCENARIOS],
)
def test_behavior(
    scenario_id,
    app,
    prog_name,
    partial,
    contains,
    excludes,
    shell_tester_factory,
):
    tester = shell_tester_factory(app, prog_name)

    # Path-completion scenarios opt in with ``contains=None``. Different
    # shells format file completions differently (fish: full path, bash:
    # basename, zsh: a mix), so exact content checks would be brittle —
    # we only assert the driver returns a non-empty result with a real
    # file in cwd.
    if contains is None:
        import os
        import tempfile
        from pathlib import Path as _Path

        with tempfile.TemporaryDirectory() as tmpdir:
            (_Path(tmpdir) / "sample.txt").write_text("x")
            cwd = _Path.cwd()
            try:
                os.chdir(tmpdir)
                results = tester.get_completions(partial)
            finally:
                os.chdir(cwd)
        assert results, f"[{scenario_id}] expected some file-completion result for {partial!r}, got nothing"
        return

    results = tester.get_completions(partial)

    missing = contains - set(results)
    assert not missing, (
        f"[{scenario_id}] expected completions {missing} missing from shell output. partial={partial!r} got={results!r}"
    )

    leaked = excludes & set(results)
    assert not leaked, (
        f"[{scenario_id}] completions {leaked} should NOT have been offered. partial={partial!r} got={results!r}"
    )



"""Tests for cyclopts.loader module."""

import sys
import textwrap
from pathlib import Path
from unittest.mock import patch

import pytest

from cyclopts.loader import load_app_from_script


@pytest.fixture
def sample_app_script(tmp_path):
    """Create a sample script with a Cyclopts app."""
    script = tmp_path / "sample_app.py"
    script.write_text(
        """\
from cyclopts import App

app = App(name="sample")

@app.default
def main(value: str = "default"):
    return value
"""
    )
    return script


@pytest.fixture
def multi_app_script(tmp_path):
    """Create a script with multiple Cyclopts apps."""
    script = tmp_path / "multi_app.py"
    script.write_text(
        """\
from cyclopts import App

app1 = App(name="first")
app2 = App(name="second")

@app1.default
def main1():
    return "app1"

@app2.default
def main2():
    return "app2"
"""
    )
    return script


def test_load_basic_script(sample_app_script):
    """Test loading a basic script with one app."""
    app, name = load_app_from_script(sample_app_script)
    assert name == "app"
    assert "sample" in app.name


def test_load_with_explicit_app_name(multi_app_script):
    """Test loading with explicit app name using : notation."""
    app, name = load_app_from_script(f"{multi_app_script}:app1")
    assert name == "app1"
    assert "first" in app.name

    app, name = load_app_from_script(f"{multi_app_script}:app2")
    assert name == "app2"
    assert "second" in app.name


def test_load_with_path_object(sample_app_script):
    """Test loading with Path object."""
    app, name = load_app_from_script(Path(sample_app_script))
    assert name == "app"
    assert "sample" in app.name


def test_load_with_path_object_and_app_name(multi_app_script):
    """Test loading with Path object using : notation."""
    # Construct a Path object that contains the :app_name notation
    path_with_app = Path(f"{multi_app_script}:app1")
    app, name = load_app_from_script(path_with_app)
    assert name == "app1"
    assert "first" in app.name

    path_with_app = Path(f"{multi_app_script}:app2")
    app, name = load_app_from_script(path_with_app)
    assert name == "app2"
    assert "second" in app.name


def test_nonexistent_file(tmp_path):
    """Test error handling for nonexistent file."""
    nonexistent = tmp_path / "nonexistent.py"
    with pytest.raises(SystemExit):
        load_app_from_script(nonexistent)


def test_non_python_file(tmp_path):
    """Test error handling for non-Python file."""
    txt_file = tmp_path / "not_python.txt"
    txt_file.write_text("This is not Python")
    with pytest.raises(SystemExit):
        load_app_from_script(txt_file)


def test_script_without_app(tmp_path):
    """Test error handling when script has no App objects."""
    script = tmp_path / "no_app.py"
    script.write_text("x = 42")
    with pytest.raises(SystemExit):
        load_app_from_script(script)


def test_invalid_app_name(sample_app_script):
    """Test error handling for invalid app name."""
    with pytest.raises(SystemExit):
        load_app_from_script(f"{sample_app_script}:nonexistent")


def test_not_an_app_object(tmp_path):
    """Test error when specified object is not an App."""
    script = tmp_path / "not_app.py"
    script.write_text(
        """\
not_an_app = "I am a string"
"""
    )
    with pytest.raises(SystemExit):
        load_app_from_script(f"{script}:not_an_app")


def test_multiple_apps_without_specification(multi_app_script):
    """Test error when multiple apps exist without specification."""
    with pytest.raises(SystemExit):
        load_app_from_script(multi_app_script)


def test_app_execution_suppressed(tmp_path):
    """Test that app() calls are suppressed during loading."""
    script = tmp_path / "executes.py"
    script.write_text(
        """\
from cyclopts import App

app = App()
executed = False

@app.default
def main():
    global executed
    executed = True
    return "executed"

# This call should be suppressed
result = app()
"""
    )
    app, _ = load_app_from_script(script)
    assert app is not None


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-specific test")
def test_windows_path_with_backslashes(tmp_path):
    """Test loading script with Windows-style backslashes."""
    script = tmp_path / "windows_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="windows")
"""
    )
    windows_path = str(script).replace("/", "\\")
    app, name = load_app_from_script(windows_path)
    assert name == "app"
    assert "windows" in app.name


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-specific test")
def test_windows_path_with_drive_letter(tmp_path):
    """Test loading script with Windows drive letter."""
    script = tmp_path / "drive_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="drive")
"""
    )
    absolute_path = script.resolve()
    app, name = load_app_from_script(str(absolute_path))
    assert name == "app"
    assert "drive" in app.name


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-specific test")
def test_windows_mixed_path_separators(tmp_path):
    """Test loading script with mixed forward/backslashes."""
    subdir = tmp_path / "subdir"
    subdir.mkdir()
    script = subdir / "mixed_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="mixed")
"""
    )
    mixed_path = str(tmp_path).replace("/", "\\") + "/subdir/mixed_app.py"
    app, name = load_app_from_script(mixed_path)
    assert name == "app"
    assert "mixed" in app.name


@pytest.mark.skipif(sys.platform != "win32", reason="Windows-specific test")
def test_windows_path_with_app_name_notation(tmp_path):
    r"""Test Windows path with :app_name notation (e.g., ``C:\path\to\file.py:app``)."""
    script = tmp_path / "windows_multi_app.py"
    script.write_text(
        textwrap.dedent(
            """\
            from cyclopts import App

            app1 = App(name="first")
            app2 = App(name="second")

            @app1.default
            def main1():
                return "app1"

            @app2.default
            def main2():
                return "app2"
            """
        )
    )
    # Test with absolute Windows path including drive letter and :app_name
    absolute_path = script.resolve()
    app, name = load_app_from_script(f"{absolute_path}:app1")
    assert name == "app1"
    assert "first" in app.name

    app, name = load_app_from_script(f"{absolute_path}:app2")
    assert name == "app2"
    assert "second" in app.name


@pytest.mark.skipif(sys.platform == "win32", reason="POSIX-specific test")
def test_posix_path_resolution(tmp_path):
    """Test POSIX path resolution."""
    script = tmp_path / "posix_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="posix")
"""
    )
    app, name = load_app_from_script(script)
    assert name == "app"
    assert "posix" in app.name


def test_path_resolution_with_symlinks(tmp_path):
    """Test loading script through symlink."""
    script = tmp_path / "original.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="original")
"""
    )
    link = tmp_path / "link.py"
    try:
        link.symlink_to(script)
        app, name = load_app_from_script(link)
        assert name == "app"
        assert "original" in app.name
    except (OSError, NotImplementedError):
        pytest.skip("Symlinks not supported on this platform")


def test_relative_path_resolution(tmp_path, monkeypatch):
    """Test loading script with relative path."""
    script = tmp_path / "relative_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="relative")
"""
    )
    monkeypatch.chdir(tmp_path)
    app, name = load_app_from_script("relative_app.py")
    assert name == "app"
    assert "relative" in app.name


def test_path_with_spaces(tmp_path):
    """Test loading script from path with spaces."""
    subdir = tmp_path / "path with spaces"
    subdir.mkdir()
    script = subdir / "spaces app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="spaces")
"""
    )
    app, name = load_app_from_script(script)
    assert name == "app"
    assert "spaces" in app.name


@patch.dict("os.environ", {"USERPROFILE": "/fake/windows/home"}, clear=False)
def test_windows_userprofile_in_path(tmp_path):
    """Test that Windows USERPROFILE environment variable doesn't break path resolution."""
    script = tmp_path / "userprofile_app.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="userprofile")
"""
    )
    app, name = load_app_from_script(script)
    assert name == "app"
    assert "userprofile" in app.name


def test_script_with_syntax_error(tmp_path):
    """Test handling of script with syntax errors."""
    script = tmp_path / "syntax_error.py"
    script.write_text("this is not valid python @@@@")
    with pytest.raises(SyntaxError):
        load_app_from_script(script)


def test_script_with_import_error(tmp_path):
    """Test handling of script with import errors."""
    script = tmp_path / "import_error.py"
    script.write_text(
        """\
from cyclopts import App
import nonexistent_module

app = App()
"""
    )
    with pytest.raises(ModuleNotFoundError):
        load_app_from_script(script)


def test_unicode_in_path(tmp_path):
    """Test loading script from path with Unicode characters."""
    subdir = tmp_path / "ディレクトリ"
    subdir.mkdir()
    script = subdir / "アプリ.py"
    script.write_text(
        """\
from cyclopts import App
app = App(name="unicode")
"""
    )
    app, name = load_app_from_script(script)
    assert name == "app"
    assert "unicode" in app.name



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
import importlib
import inspect
import sys
from datetime import date
from pathlib import Path

import git
from sphinx.application import Sphinx
from sphinx.ext.autodoc import Options

from cyclopts import __version__

sys.path.insert(0, str(Path("../..").absolute()))


git_repo = git.Repo(".", search_parent_directories=True)  # type: ignore[reportPrivateImportUsage]
git_commit = git_repo.head.commit

# -- Project information -----------------------------------------------------

project = "cyclopts"
copyright = f"{date.today().year}, Brian Pugh"
author = "Brian Pugh"

# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx_rtd_theme",
    "sphinx_rtd_dark_mode",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
    "cyclopts.sphinx_ext",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

smartquotes = False

# user starts in light mode
default_dark_mode = False

# Myst
myst_enable_extensions = [
    "linkify",
]

# Intersphinx
intersphinx_mapping = {
    "pydantic": ("https://docs.pydantic.dev/latest/", None),
    "python": ("https://docs.python.org/3", None),
    "rich": ("https://rich.readthedocs.io/en/stable/", None),
    "typing_extensions": ("https://typing-extensions.readthedocs.io/en/latest/", None),
    "pytest": ("https://docs.pytest.org/en/latest", None),
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Autodoc
autodoc_default_options = {
    "member-order": "bysource",
    "undoc-members": False,
    "exclude-members": "__weakref__",
}
autoclass_content = "both"

# LinkCode
code_url = f"https://github.com/BrianPugh/cyclopts/blob/{git_commit}"


def linkcode_resolve(domain, info):
    """Link code to github.

    Modified from:
        https://github.com/python-websockets/websockets/blob/778a1ca6936ac67e7a3fe1bbe585db2eafeaa515/docs/conf.py#L100-L134
    """
    # Non-linkable objects from the starter kit in the tutorial.
    if domain == "js":
        return

    if domain != "py":
        raise ValueError("expected only Python objects")

    if not info.get("module"):
        # Documented via py:function::
        return

    mod = importlib.import_module(info["module"])
    if "." in info["fullname"]:
        objname, attrname = info["fullname"].split(".")
        obj = getattr(mod, objname)
        try:
            # object is a method of a class
            obj = getattr(obj, attrname)
        except AttributeError:
            # object is an attribute of a class
            return None
    else:
        obj = getattr(mod, info["fullname"])

    try:
        file = inspect.getsourcefile(obj)
        lines = inspect.getsourcelines(obj)
    except TypeError:
        # e.g. object is a typing.Union
        return None
    if file is None:
        return None
    file = Path(file).resolve().relative_to(git_repo.working_dir)
    if file.parts[0] != "cyclopts":
        # e.g. object is a typing.NewType
        return None
    start, end = lines[1], lines[1] + len(lines[0]) - 1

    return f"{code_url}/{file}#L{start}-L{end}"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_title = project
html_logo = "../../assets/logo_512w.png"
html_favicon = "../../assets/favicon-192.png"

html_theme_options = {
    "logo_only": True,
    "version_selector": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    "vcs_pageview_mode": "",
    "style_nav_header_background": "#F7E5B9",
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}

html_context = {
    # Github options
    "display_github": True,
    "github_user": "BrianPugh",
    "github_repo": "cyclopts",
    "github_version": "main",
    "conf_py_path": "/docs/source/",
}

html_css_files = [
    "custom.css",
]


# --- Other Custom Stuff


def simplify_exception_signature(
    app: Sphinx, what: str, name: str, obj, options: Options, signature, return_annotation
):
    # Check if the object is an exception and modify the signature
    if what == "exception" and isinstance(obj, type) and issubclass(obj, BaseException):
        return ("", None)  # Return an empty signature and no return annotation


def remove_attrs_methods(app, what, name, obj, options, lines):
    lines[:] = [line for line in lines if not line.startswith("Method generated by attrs for")]


def setup(app: Sphinx):
    app.connect("autodoc-process-signature", simplify_exception_signature)
    app.connect("autodoc-process-docstring", remove_attrs_methods)



"""Tests for enum.Flag and enum.IntFlag support in cyclopts."""

from dataclasses import dataclass
from enum import Flag, IntFlag, auto
from textwrap import dedent
from typing import Annotated

import pytest

from cyclopts import MissingArgumentError, Parameter
from cyclopts.exceptions import CoercionError, UnknownOptionError


class Permission(Flag):
    """File permissions as bit flags."""

    READ = auto()
    """Enable read permissions."""
    WRITE = auto()
    """Enable write permissions."""
    EXECUTE = auto()
    """Enable execute permissions."""


class HttpStatus(IntFlag):
    INFORMATIONAL = 100
    SUCCESS = 200
    REDIRECTION = 300
    CLIENT_ERROR = 400
    SERVER_ERROR = 500


@dataclass
class User:
    name: str
    perms: Permission = Permission.READ


def test_flag_as_boolean_flags(app, assert_parse_args):
    """Test that Flag enum members are exposed as boolean CLI flags."""

    @app.default
    def main(perms: Permission = Permission.READ):
        pass

    assert_parse_args(main, "--perms.read", perms=Permission.READ)
    assert_parse_args(main, "--perms.write", perms=Permission.WRITE)
    assert_parse_args(main, "--perms.read --perms.write", perms=Permission.READ | Permission.WRITE)


def test_int_flag_as_boolean_flags(app, assert_parse_args):
    """Test that IntFlag enum members are exposed as boolean CLI flags."""

    @app.default
    def main(error_on: HttpStatus | None = None):
        pass

    assert_parse_args(main, "")
    assert_parse_args(main, "--error-on.success", error_on=HttpStatus.SUCCESS)


@pytest.mark.parametrize(
    "command",
    ["--perms read --perms write", "read write"],
)
def test_flag_with_list_str_input(app, assert_parse_args, command):
    """Test that Flag enum members are exposed as boolean CLI flags."""

    @app.default
    def main(perms: Permission = Permission.READ):
        pass

    assert_parse_args(main, command, Permission.READ | Permission.WRITE)


def test_flag_unknown_member_str(app):
    @app.default
    def main(perms: Permission = Permission.READ):
        pass

    with pytest.raises(CoercionError):
        app("foo", exit_on_error=False)


def test_flag_unknown_member_option(app):
    @app.default
    def main(perms: Permission = Permission.READ):
        pass

    with pytest.raises(UnknownOptionError):
        app("--perms.foo", exit_on_error=False)


def test_flag_as_boolean_flags_star_name(app, assert_parse_args, console):
    """Test that Flag enum members are exposed as boolean CLI flags."""

    @app.default
    def main(perms: Annotated[Permission, Parameter(name="*", negative_bool="")] = Permission.READ):
        """Manage file permissions."""
        pass

    # Test individual flags
    assert_parse_args(main, "--read", Permission.READ)
    assert_parse_args(main, "--write", Permission.WRITE)
    assert_parse_args(main, "--execute", Permission.EXECUTE)

    # Test multiple flags combined
    assert_parse_args(main, "--read --write", Permission.READ | Permission.WRITE)
    assert_parse_args(main, "--read --write --execute", Permission.READ | Permission.WRITE | Permission.EXECUTE)

    # Test help output
    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()
    expected = dedent(
        """\
        Usage: test_enum_flag [OPTIONS]

        Manage file permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ --read     Enable read permissions. [default: False]               │
        │ --write    Enable write permissions. [default: False]              │
        │ --execute  Enable execute permissions. [default: False]            │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )
    assert expected == actual


def test_flag_with_default_value(app, assert_parse_args):
    """Test Flag enum with non-zero default value."""

    @app.default
    def main(perms: Permission = Permission.READ | Permission.WRITE):
        pass

    assert_parse_args(main, "")


def test_flag_with_other_parameters(app, assert_parse_args):
    """Test Flag enum alongside other parameters."""

    @app.default
    def main(name: str, perms: Permission = Permission.READ, verbose: bool = False):
        pass

    # Test with positional and flag parameters
    assert_parse_args(main, "test.txt --perms.read --perms.write", "test.txt", Permission.READ | Permission.WRITE)

    # Test with all parameter types
    assert_parse_args(main, "test.txt --perms.execute --verbose", "test.txt", Permission.EXECUTE, True)


def test_flag_star_name_with_other_parameters(app, assert_parse_args):
    """Test Flag enum with star name alongside other parameters."""

    @app.default
    def main(
        name: str,
        perms: Annotated[Permission, Parameter(name="*")] = Permission.READ,
        verbose: bool = False,
    ):
        pass

    # Test star name expansion with other parameters
    assert_parse_args(main, "test.txt --read --write --verbose", "test.txt", Permission.READ | Permission.WRITE, True)


def test_flag_no_flags_provided(app):
    """Test behavior when no flags are provided but parameter is required."""

    @app.default
    def main(perms: Permission):  # No default value
        pass

    with pytest.raises(MissingArgumentError):
        app([], exit_on_error=False)


def test_flag_help_shows_member_docstrings(app, console):
    """Test that Flag enum member docstrings appear in help output."""

    @app.default
    def main(perms: Permission = Permission.READ):
        """Manage file permissions."""
        pass

    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()

    expected = dedent(
        """\
        Usage: test_enum_flag [OPTIONS]

        Manage file permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ --perms.read          Enable read permissions. [default: False]    │
        │   --perms.no-read                                                  │
        │ --perms.write         Enable write permissions. [default: False]   │
        │   --perms.no-write                                                 │
        │ --perms.execute       Enable execute permissions. [default: False] │
        │   --perms.no-execute                                               │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )
    assert expected == actual


def test_flag_help_star_name_shows_member_docstrings(app, console):
    """Test that Flag enum member docstrings appear with star name expansion."""

    @app.default
    def main(perms: Annotated[Permission, Parameter(name="*")] = Permission.READ):
        """Manage file permissions."""
        pass

    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()
    expected = dedent(
        """\
        Usage: test_enum_flag [OPTIONS]

        Manage file permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ --read --no-read        Enable read permissions. [default: False]  │
        │ --write --no-write      Enable write permissions. [default: False] │
        │ --execute --no-execute  Enable execute permissions. [default:      │
        │                         False]                                     │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )

    assert expected == actual


def test_flag_in_dataclass(app, assert_parse_args):
    """Test Flag enum field in a dataclass."""

    @app.default
    def main(user: User):
        pass

    # Test with name only (should use default perms)
    assert_parse_args(main, "Alice", User("Alice", Permission.READ))

    # Test with name and flag permissions
    assert_parse_args(
        main, "Bob --user.perms.write --user.perms.execute", User("Bob", Permission.WRITE | Permission.EXECUTE)
    )


def test_flag_in_dataclass_positionally(app, assert_parse_args):
    """Test Flag enum field positionally in a dataclass."""

    @app.default
    def main(user: User):
        pass

    assert_parse_args(main, "Bob write execute", User("Bob", Permission.WRITE | Permission.EXECUTE))


def test_flag_in_dataclass_help(app, console):
    """Test help output for dataclass with Flag enum field."""

    @app.default
    def main(user: User):
        """Create a user with permissions."""
        pass

    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()
    expected = dedent(
        """\
        Usage: test_enum_flag [OPTIONS] USER.NAME

        Create a user with permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ *  USER.NAME --user.name      [required]                           │
        │    --user.perms.read          Enable read permissions. [default:   │
        │      --user.perms.no-read     False]                               │
        │    --user.perms.write         Enable write permissions. [default:  │
        │      --user.perms.no-write    False]                               │
        │    --user.perms.execute       Enable execute permissions.          │
        │      --user.perms.no-execute  [default: False]                     │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )

    assert expected == actual


def test_flag_in_dataclass_help_no_negative(app, console):
    """Test help output for dataclass with Flag enum field."""

    @app.default
    def main(user: Annotated[User, Parameter(negative="")]):
        """Create a user with permissions."""
        pass

    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()
    expected = dedent(
        """\
        Usage: test_enum_flag [OPTIONS] USER.NAME

        Create a user with permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ *  USER.NAME --user.name  [required]                               │
        │    --user.perms.read      Enable read permissions. [default:       │
        │                           False]                                   │
        │    --user.perms.write     Enable write permissions. [default:      │
        │                           False]                                   │
        │    --user.perms.execute   Enable execute permissions. [default:    │
        │                           False]                                   │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )

    assert expected == actual


def test_flag_in_dataclass_help_no_keywords(app, assert_parse_args, console):
    """Test help output for dataclass with Flag enum field."""

    @dataclass
    class User:
        name: str
        perms: Annotated[Permission, Parameter(accepts_keys=False)] = Permission.READ

    @app.default
    def main(user: Annotated[User, Parameter(negative="")]):
        """Create a user with permissions."""
        pass

    with console.capture() as capture:
        app.help_print([], console=console)

    actual = capture.get()
    expected = dedent(
        """\
        Usage: test_enum_flag USER.NAME [ARGS]

        Create a user with permissions.

        ╭─ Commands ─────────────────────────────────────────────────────────╮
        │ --help (-h)  Display this message and exit.                        │
        │ --version    Display application version.                          │
        ╰────────────────────────────────────────────────────────────────────╯
        ╭─ Parameters ───────────────────────────────────────────────────────╮
        │ *  USER.NAME --user.name    [required]                             │
        │    USER.PERMS --user.perms  [choices: read, write, execute]        │
        │                             [default: read]                        │
        ╰────────────────────────────────────────────────────────────────────╯
        """
    )

    assert expected == actual

    assert_parse_args(main, "Bob read write", User("Bob", perms=Permission.READ | Permission.WRITE))



from enum import Enum
from pathlib import Path
from typing import Annotated, Literal

from cyclopts import App, Parameter
from cyclopts.types import ExistingFile

app_basic = App(name="basic")


@app_basic.default
def main(
    verbose: Annotated[bool, Parameter(help="Enable verbose output")] = False,
    count: Annotated[int, Parameter(help="Number of items")] = 1,
):
    """Basic app for testing."""
    pass


@app_basic.command
def deploy(
    env: Annotated[Literal["dev", "staging", "prod"], Parameter(help="Environment")],
    force: bool = False,
):
    """Deploy to environment."""
    pass


class Speed(Enum):
    FAST = "fast"
    SLOW = "slow"


app_enum = App(name="enumapp")


@app_enum.default
def enum_main(speed: Speed = Speed.FAST):
    """App with enum parameter."""
    pass


app_nested = App(name="nested")
config_app = App(name="config")


@config_app.command
def get(key: str):
    """Get config value."""
    pass


@config_app.command
def set(key: str, value: str):
    """Set config value."""
    pass


app_nested.command(config_app)


app_path = App(name="pathapp")


@app_path.default
def path_main(
    input_file: Annotated[Path, Parameter(help="Input file")],
    output: Annotated[Path | None, Parameter(help="Output file")] = None,
):
    """App with path parameters."""
    pass


app_negative = App(name="negapp")


@app_negative.default
def negative_main(
    verbose: Annotated[bool, Parameter(help="Enable verbose")] = False,
    colors: Annotated[bool, Parameter(negative="no-colors")] = True,
):
    """App with negative flags."""
    pass


app_markup = App(name="markupapp")


@app_markup.default
def markup_main(
    verbose: Annotated[bool, Parameter(help="Enable **verbose** output with `extra` details")] = False,
    mode: Annotated[
        Literal["fast", "slow"],
        Parameter(help="Choose *execution* mode: **fast** or **slow**"),
    ] = "fast",
):
    """App with **markdown** markup in help text.

    This tests that markup is properly stripped in completions.
    """
    pass


@app_markup.command
def deploy_markup(
    env: Annotated[str, Parameter(help="Target `environment` like **dev** or **prod**")],
):
    """Deploy to `environment`."""
    pass


app_rst = App(name="rstapp", help_format="rst")


@app_rst.default
def rst_main(
    verbose: Annotated[bool, Parameter(help="Enable **verbose** output with ``code`` samples")] = False,
    mode: Annotated[
        Literal["fast", "slow"],
        Parameter(help="Choose *execution* mode: **fast** or **slow**"),
    ] = "fast",
):
    """App with **RST** markup in help text.

    This tests that RST markup is properly stripped in completions.
    """
    pass


app_positional_literal = App(name="poslit")


@app_positional_literal.command
def command(param: Literal["foo", "bar", "baz"], /):
    """Simple command with positional literal.

    Parameters
    ----------
    param : Literal["foo", "bar", "baz"]
        Literal param.
    """
    pass


app_multiple_positionals = App(name="multipos")


@app_multiple_positionals.command
def command_multi(
    first: Literal["red", "blue"],
    second: Literal["cat", "dog"],
    /,
):
    """Command with multiple positionals with distinct choices.

    Parameters
    ----------
    first : Literal["red", "blue"]
        Color choice.
    second : Literal["cat", "dog"]
        Animal choice.
    """
    pass


app_deploy = App(name="deploy")


@app_deploy.command
def deploy_project(
    project: Literal["web", "api", "worker"],
    /,
    *,
    environment: Literal["dev", "staging", "prod"],
    branch: str = "main",
):
    """Deploy a project to an environment.

    Parameters
    ----------
    project : Literal["web", "api", "worker"]
        Project to deploy.
    environment : Literal["dev", "staging", "prod"]
        Target environment.
    branch : str
        Git branch.
    """
    pass


app_positional_path = App(name="pathpos")


@app_positional_path.command
def process(
    input_file: Path,
    /,
):
    """Process a file.

    Parameters
    ----------
    input_file : Path
        Input file to process.
    """
    pass


app_list_path = App(name="listpath")


@app_list_path.default
def list_path_main(
    files: Annotated[list[Path], Parameter(help="List of files")],
):
    """App with list[Path] parameter."""
    pass


app_list_annotated_path = App(name="listannotatedpath")


@app_list_annotated_path.default
def list_annotated_path_main(
    files: list[ExistingFile],
):
    """App with list[ExistingFile] parameter."""
    pass


app_disabled_negative = App(name="disabledneg", default_parameter=Parameter(negative=""))


@app_disabled_negative.command
def build(param: list[Literal["apple", "banana", "cherry"]]):
    """Build command with list parameter."""
    pass


app_three_positionals = App(name="multipos3")


@app_three_positionals.command
def command_multi3(
    first: Literal["red", "blue"],
    second: Literal["cat", "dog"],
    third: Literal["small", "large"],
    /,
):
    """Command with three positional literals.

    Parameters
    ----------
    first : Literal["red", "blue"]
        Color choice.
    second : Literal["cat", "dog"]
        Animal choice.
    third : Literal["small", "large"]
        Size choice.
    """
    pass


# Two iterable positionals back-to-back. Exercises the "first iterable wins
# the rest position" rule that prevents two rest specs from being emitted
# (regressions here would surface as either invalid scripts or wrong
# completions at the rest position).
app_two_iterables = App(name="twoiter")


@app_two_iterables.command
def collect(
    paths: list[Path],
    tags: list[str],
):
    """Collect paths and tags.

    Parameters
    ----------
    paths : list[Path]
        Paths to collect.
    tags : list[str]
        Optional tags.
    """
    pass



import inspect
from collections import namedtuple
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, Union

import pytest

from cyclopts.annotations import contains_hint, get_hint_name, is_iterable_type, resolve


def test_resolve_annotated():
    type_ = Annotated[Literal["foo", "bar"], "fizz"]
    res = resolve(type_)
    assert res == Literal["foo", "bar"]


def test_resolve_empty():
    res = resolve(inspect.Parameter.empty)
    assert res is str


def test_get_hint_name_string():
    assert get_hint_name("str") == "str"


def test_get_hint_name_any():
    assert get_hint_name(Any) == "Any"


def test_get_hint_name_union():
    assert get_hint_name(Union[int, str]) == "int|str"


def test_get_hint_name_class_with_name():
    class TestClass:
        pass

    assert get_hint_name(TestClass) == "TestClass"


def test_get_hint_name_typing_with_name():
    assert get_hint_name(list) == "list"


def test_get_hint_name_generic_type():
    assert get_hint_name(list[int]) == "list[int]"


def test_get_hint_name_nested_generic_type():
    assert get_hint_name(dict[str, list[int]]) == "dict[str, list[int]]"


def test_get_hint_name_optional_type():
    assert get_hint_name(Optional[int]) == "int|None"


def test_get_hint_name_namedtuple():
    TestTuple = namedtuple("TestTuple", ["field1", "field2"])
    assert get_hint_name(TestTuple) == "TestTuple"


def test_get_hint_name_complex_union():
    complex_type = Union[int, str, list[dict[str, Any]]]
    assert get_hint_name(complex_type) == "int|str|list[dict[str, Any]]"


def test_get_hint_name_fallback_str():
    class NoNameClass:
        def __str__(self):
            return "NoNameClass"

    assert get_hint_name(NoNameClass()) == "NoNameClass"


class CustomStr(str):
    """Dummy subclass of ``str``."""


@pytest.mark.parametrize(
    "hint,target_type,expected",
    [
        (str, str, True),
        (CustomStr, str, True),
        (Union[int, str], str, True),
        (Annotated[int | str, 1], str, True),
        (Annotated[Annotated[int, 1] | Annotated[str, 1], 1], str, True),
        (int, str, False),
    ],
)
def test_contains_hint(hint, target_type, expected):
    assert contains_hint(hint, target_type) == expected


@pytest.mark.parametrize(
    "hint,expected",
    [
        # Basic collection types
        (list[str], True),
        (set[int], True),
        (tuple[str, ...], True),
        (frozenset[str], True),
        # Abstract collection types
        (Sequence[str], True),
        (Iterable[str], True),
        # Annotated wrappers
        (Annotated[list[str], "metadata"], True),
        # Optional wrappers
        (Optional[list[str]], True),
        # Nested Annotated inside collection
        (list[Annotated[Path, "metadata"]], True),
        # Non-iterable types
        (str, False),
        (int, False),
        (Path, False),
        (dict[str, str], True),
        (Annotated[str, "metadata"], False),
        (Optional[str], False),
    ],
)
def test_is_iterable_type(hint, expected):
    assert is_iterable_type(hint) == expected


def test_get_annotated_discriminator_rejects_non_annotated():
    from cyclopts.annotations import get_annotated_discriminator

    class Decoy:
        discriminator = "type"

    assert get_annotated_discriminator(list[Decoy]) is None
    assert get_annotated_discriminator(int) is None


def test_get_annotated_discriminator_positive():
    from types import SimpleNamespace

    from cyclopts.annotations import get_annotated_discriminator

    assert get_annotated_discriminator(Annotated[int, SimpleNamespace(discriminator="type")]) == "type"
